# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Get cool commit messages.

Based on: https://github.com/ngerakines/commitment
"""

from __future__ import annotations

import logging
import random
from dataclasses import dataclass
from datetime import datetime
from typing import Final, TypeAlias

import emoji
from tornado.httpclient import AsyncHTTPClient
from tornado.web import HTTPError

from ..utils.data_parsing import parse_args
from ..utils.request_handler import APIRequestHandler
from ..utils.utils import ModuleInfo

LOGGER: Final = logging.getLogger(__name__)


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=((r"/api/commitment", CommitmentAPI),),
        name="Commitment",
        short_name="Commitment",
        description="Zeige gute Commit-Nachrichten an.",
        path="/api/commitment",
        aliases=(),
        sub_pages=(),
        keywords=(),
        hidden=True,
    )


Commit: TypeAlias = tuple[datetime, str]
Commits: TypeAlias = dict[str, Commit]
COMMIT_DATA: dict[str, Commits] = {}


async def get_commit_data(commitment_uri: str) -> Commits:
    """Get data from URI."""
    if commitment_uri in COMMIT_DATA:
        return COMMIT_DATA[commitment_uri]
    file_content: bytes
    if commitment_uri.startswith(("https://", "http://")):
        file_content = (await AsyncHTTPClient().fetch(commitment_uri)).body
    else:
        if commitment_uri.startswith("file:///"):
            commitment_uri = commitment_uri.removeprefix("file://")
        with open(commitment_uri, "rb") as file:
            file_content = file.read()

    data: Commits = {}

    for line in file_content.decode("UTF-8").split("\n"):
        if not line:
            continue
        hash_, date, msg = line.split(" ", 2)
        data[hash_] = (datetime.utcfromtimestamp(int(date)), msg)

    COMMIT_DATA[commitment_uri] = data
    return data


@dataclass(slots=True)
class Arguments:
    """The arguments for the commitment API."""

    hash: str | None = None
    require_emoji: bool = False


class CommitmentAPI(APIRequestHandler):
    """The request handler for the commitment API."""

    POSSIBLE_CONTENT_TYPES = (
        "text/plain",
        *APIRequestHandler.POSSIBLE_CONTENT_TYPES,
    )

    @parse_args(type_=Arguments)
    async def get(self, *, args: Arguments, head: bool = False) -> None:
        """Handle GET requests to the API."""
        # pylint: disable=unused-argument
        try:
            data = await get_commit_data(self.settings["COMMITMENT_URI"])
        except Exception as exc:
            raise HTTPError(503) from exc

        if args.hash is None:
            return await self.write_commit(
                *random.choice(
                    [
                        (com, (_, msg))
                        for com, (_, msg) in data.items()
                        if not args.require_emoji or emoji.emoji_count(msg)
                    ]
                )
            )

        if len(args.hash) + 2 == 42:
            if args.hash in data:
                return await self.write_commit(args.hash, data[args.hash])
            raise HTTPError(404)

        if len(args.hash) + 1 >= 42:
            raise HTTPError(404)

        results = [
            item
            for item in data.items()
            if item[0].startswith(args.hash)
            if not args.require_emoji or emoji.emoji_count(item[1][1])
        ]

        if not results:
            raise HTTPError(404)

        results.sort(key=lambda m: m[1][0])

        return await self.write_commit(*results[0])

    async def write_commit(self, hash_: str, commit: Commit) -> None:
        """Write the commit data."""
        self.set_header("X-Commit-Hash", hash_)

        if self.content_type == "text/plain":
            return await self.finish(commit[1])

        return await self.finish_dict(
            hash=hash_,
            commit_message=commit[1],
            permalink=self.fix_url("/api/commitment", hash=hash_),
            date=commit[0],
        )



#!/usr/bin/env python3

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The client for the backdoor API of the website."""

from __future__ import annotations

import ast
import asyncio
import contextlib
import http.client
import io
import os
import pickle  # nosec: B403
import pydoc
import re  # pylint: disable=preferred-module
import socket
import sys
import time
import traceback
import uuid
from base64 import b64encode
from collections.abc import Callable, Iterable, MutableMapping
from textwrap import dedent
from types import EllipsisType
from typing import Any, Required, TypeAlias, TypedDict, cast
from urllib.parse import SplitResult, quote, quote_plus, urlsplit

with contextlib.suppress(ModuleNotFoundError):
    # pylint: disable=shadowed-import
    import dill as pickle  # type: ignore[import-untyped, no-redef]  # noqa: F811, B950  # nosec: B403

try:
    import hy  # type: ignore[import-untyped]
except ModuleNotFoundError:
    hy = None  # pylint: disable=invalid-name

try:
    import idna
except ModuleNotFoundError:
    idna = None  # type: ignore[assignment]  # pylint: disable=invalid-name

try:
    import socks  # type: ignore[import-untyped]
except ModuleNotFoundError:
    socks = None  # pylint: disable=invalid-name

try:
    import uvloop
except ModuleNotFoundError:
    pass
else:
    uvloop.install()


FLUFL = True

E = eval(  # pylint: disable=eval-used  # nosec: B307
    "eval(repr((_:=[],_.append(_))[0]))[0][0]"
)

ErrorTuple: TypeAlias = tuple[int, str]


class Response(TypedDict):  # noqa: D101
    # pylint: disable=missing-class-docstring
    success: bool | EllipsisType
    output: None | str
    result: None | tuple[str, None | bytes] | SystemExit


class Proxy(TypedDict, total=False):  # noqa: D101
    # pylint: disable=missing-class-docstring
    type: Required[int]
    addr: Required[str]
    port: int
    rdns: bool
    username: str
    password: str


async def create_socket(
    addr: str,
    port: int | str,
    proxy: None | Proxy,
) -> socket.socket:
    """Create a socket (optionally with a proxy)."""
    # pylint: disable=too-complex
    if proxy is not None and not socks:
        raise NotImplementedError("PySocks is required for proxy support")
    loop = asyncio.get_running_loop()
    address_infos = await loop.getaddrinfo(
        addr,
        port,
        # PySocks doesn't support AF_INET6
        family=0 if proxy is None else socket.AF_INET,
        type=socket.SOCK_STREAM,
    )
    if not address_infos:
        raise OSError("getaddrinfo() returned empty list")
    exceptions = []
    for address_info in address_infos:
        sock = None
        try:
            sock = cast(
                socket.socket,
                socks.socksocket(
                    address_info[0], address_info[1], address_info[2]
                )
                if proxy is not None
                else socket.socket(
                    address_info[0], address_info[1], address_info[2]
                ),
            )
            sock.setblocking(False)
            if proxy is not None:
                sock.set_proxy(  # type: ignore[attr-defined]
                    proxy["type"],
                    proxy["addr"],
                    proxy.get("port"),
                    proxy.get("rdns", True),
                    proxy.get("username"),
                    proxy.get("password"),
                )
            await loop.sock_connect(sock, address_info[4])
            return sock
        except OSError as exc:
            if sock is not None:
                sock.close()
            exceptions.append(exc)
            continue
        except BaseException:
            if sock is not None:
                sock.close()
            raise
    if len(exceptions) == 1:
        raise exceptions[0]
    model = str(exceptions[0])
    if all(str(exc) == model for exc in exceptions):
        raise exceptions[0]
    raise OSError(
        f"Multiple exceptions: {', '.join(str(exc) for exc in exceptions)}"
    )


async def request(  # noqa: C901
    method: str,
    url: str | SplitResult,
    headers: None | MutableMapping[str, str] = None,
    body: None | bytes | bytearray | Iterable[bytes | bytearray] | str = None,
    proxy: None | Proxy = None,
) -> tuple[int, dict[str, str], bytes]:
    """Insanely awesome HTTP client."""
    # pylint: disable=line-too-long, too-complex
    # pylint: disable=too-many-branches, too-many-locals, while-used
    if isinstance(url, str):
        url = urlsplit(url)
    if url.scheme not in {"", "http", "https"}:
        raise ValueError(f"Unsupported scheme: {url.scheme}")
    if not url.hostname:
        raise ValueError("URL has no hostname")
    if headers is None:
        headers = {}
    if isinstance(body, str):
        body = body.encode("UTF-8")
    if isinstance(body, memoryview):
        body = body.tobytes()  # type: ignore[unreachable]
    if isinstance(body, Iterable) and not isinstance(body, (bytes, bytearray)):
        body = b"".join(body)  # type: ignore[arg-type]
    https = url.scheme == "https"
    header_names = [x.strip().title() for x in headers.keys()]
    if "Host" not in header_names:
        host: None | str = None
        if idna:  # type: ignore[truthy-bool]
            with contextlib.suppress(idna.core.InvalidCodepoint):
                host = idna.encode(url.hostname).decode("ASCII")
                host = f"{host}:{url.port}" if url.port else host
        if not host:
            host = url.netloc.encode("IDNA").decode("ASCII")
        headers["Host"] = host
    if body and "Content-Length" not in header_names:
        headers["Content-Length"] = str(len(body))
    sock = await create_socket(
        url.hostname,
        url.port or ("https" if https else "http"),
        proxy,
    )
    reader, writer = await asyncio.open_connection(
        sock=sock,
        ssl=https,
        server_hostname=url.hostname if https else None,
    )
    writer.write(
        (
            method
            + " "
            + (quote(url.path) or "/")
            + ("?" + quote_plus(url.query) if url.query else "")
            + " HTTP/1.0\r\n"
        ).encode("ASCII")
        + "\r\n".join(
            [f"{key}:{value}" for key, value in headers.items()] + [""]
        ).encode("latin-1")
        + b"\r\n"
    )
    if body:
        writer.write(body)
    await writer.drain()
    del headers
    e, data = E, b""
    while chunk := await reader.read():
        if b"\r\n\r\n" in (data := data + chunk) and e is E:
            e, data = data.split(b"\r\n\r\n", 1)
            status, o = re.match(r"HTTP/.+? (\d+).*?\r\n(.*)", e.decode("LATIN-1"), 24).groups()  # type: ignore[union-attr]  # noqa: B950
            headers = dict(re.match(r"([^\s]+):\s*(.+?)\s*$", x, 24).groups() for x in o.split("\r\n"))  # type: ignore[union-attr, misc]  # noqa: B950
    writer.close()
    await writer.wait_closed()
    if "status" not in locals():
        raise AssertionError("No HTTP response received")
    return int(status), headers, data  # type: ignore[possibly-undefined]


def detect_mode(code: str) -> str:
    """Detect which mode needs to be used."""
    import __future__  # pylint: disable=import-outside-toplevel

    flags: int = ast.PyCF_ONLY_AST
    if FLUFL:
        flags |= __future__.barry_as_FLUFL.compiler_flag

    try:
        compile(code, "", "eval", flags, cast(bool, 0x5F3759DF), 0)
        return "eval"
    except SyntaxError:
        compile(code, "", "exec", flags, cast(bool, 0x5F3759DF), 0)
        return "exec"


def send(
    url: str | SplitResult,
    key: str,
    code: str,
    mode: str = "exec",
    session: None | str = None,
    proxy: None | Proxy = None,
) -> tuple[int, dict[str, str], Response | ErrorTuple | str | None | bytes]:
    """Send code to the backdoor API."""
    # pylint: disable=too-many-arguments
    body = code.encode("UTF-8")
    if isinstance(url, str):
        url = urlsplit(url)
    if not url.path:
        url = url._replace(path="/api/backdoor")
    key = f"Bearer {b64encode(key.encode('UTF-8')).decode('ASCII')}"
    headers = {
        "Authorization": key,
        "Accept": "application/vnd.uqfoundation.dill"
        if pickle.__name__ == "dill"
        else "application/vnd.python.pickle",
        "X-Pickle-Protocol": str(pickle.HIGHEST_PROTOCOL),
        "X-Future-Feature": "annotations",
    }
    if FLUFL:
        headers["X-Future-Feature"] += ", barry_as_FLUFL"
    if session:
        headers["X-Backdoor-Session"] = session
    response = asyncio.run(
        request(
            "POST",
            url._replace(path=f"{url.path.removesuffix('/')}/{mode}"),
            headers,
            body,
            proxy,
        )
    )
    try:
        return (
            response[0],  # status
            response[1],  # header
            pickle.loads(response[2]),  # data  # nosec: B301
        )
    except pickle.UnpicklingError:
        return response


def lisp_always_active() -> bool:
    """Return True if LISP is always active."""
    return (
        hy
        and not hy.eval(
            hy.read(
                '(* (- (* (+ 0 1) 2 3 4 5) (+ 6 7 8 9 10 11)) '  # fmt: skip
                '(int (= (. (__import__ "os.path") sep) "/")))'
            )
        )
        and not int.from_bytes(
            getattr(
                os, "洀漀搀渀愀爀甀".encode("UTF-16-BE")[::-1].decode("UTF-16-BE")
            )(1),
            "big",
        )
        // (69 // 4 - 1)
    )


def run_and_print(  # noqa: C901
    url: str,
    key: str,
    code: str,
    lisp: bool = False,
    session: None | str = None,
    proxy: None | Proxy = None,
    time_requests: bool = False,
    *,
    # pylint: disable=redefined-builtin
    print: Callable[..., None] = print,
) -> None:
    """Run the code and print the output."""
    # pylint: disable=too-complex, too-many-arguments, too-many-branches
    # pylint: disable=too-many-locals, too-many-statements
    start_time = time.monotonic()
    if lisp or lisp_always_active():
        code = hy.disassemble(hy.read(code), True)
    try:
        response = send(
            url,
            key,
            code,
            detect_mode(code),
            session,
            proxy,
        )
    except SyntaxError as exc:
        print("".join(traceback.format_exception_only(exc)).strip())
        return
    if time_requests:
        took = time.monotonic() - start_time
        if took > 1:
            color = "91"  # red
        elif took > 0.5:
            color = "93"  # yellow
        else:
            color = "92"  # green
        print(f"\033[{color}mTook: {took:.3f}s\033[0m")
    status, headers, body = response  # pylint: disable=unused-variable
    if status >= 400:
        reason = (
            body[1]
            if isinstance(body, tuple)
            else http.client.responses[status]
        )
        print("\033[91m" + f"{status} {reason}" + "\033[0m")
        if isinstance(body, str):
            print("\033[91m" + body + "\033[0m")
    elif body is None:  # pylint: disable=confusing-consecutive-elif
        pass
    elif isinstance(body, dict):
        if isinstance(body["success"], bool):
            print(f"Success: {body['success']}")
        if isinstance(body["output"], str) and body["output"]:
            print("Output:")
            print(body["output"].strip())
        if isinstance(body["result"], SystemExit):
            raise body["result"]
        if isinstance(body["result"], tuple):
            if not body["success"]:
                print(body["result"][0])
                return
            result_obj: Any = None
            if isinstance(body["result"][1], bytes):
                try:
                    result_obj = pickle.loads(body["result"][1])  # nosec: B301
                except Exception:  # pylint: disable=broad-except
                    if sys.flags.dev_mode:
                        traceback.print_exc()
            if (
                isinstance(result_obj, tuple)
                and len(result_obj) == 2
                and result_obj[0] == "PagerTuple"
                and isinstance(result_obj[1], str)
            ):
                pydoc.pager(result_obj[1])
            else:
                print("Result:")
                print(body["result"][0])
    else:
        print(f"Response has unexpected type {type(body).__name__}!")
        print(body)


def shellify(code: str) -> str:
    """Modify code in a way that it gets executed in a shell."""
    if not code.startswith("!"):
        return code

    code = (
        code[1:]
        .strip()
        .replace("\\", r"\\")
        .replace("\n", r"\n")
        .replace('"', r"\"")
    )
    return f"""
async def run_shell_50821273052022fbc283():
    import asyncio
    proc = await asyncio.create_subprocess_shell(
        {code!r},
        asyncio.subprocess.DEVNULL,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )
    output, _ = await asyncio.wait_for(proc.communicate(), 60 * 60)
    if output:
        print(output.decode("UTF-8"))
await run_shell_50821273052022fbc283()
del run_shell_50821273052022fbc283
"""


def main() -> int | str:  # noqa: C901
    """Parse arguments, load the config and start the backdoor client."""
    # pylint: disable=too-complex, too-many-branches
    # pylint: disable=too-many-locals, too-many-statements
    if "--help" in sys.argv or "-h" in sys.argv:
        return dedent(
            """\
            Accepted arguments:

                --dev              use a separate config for a local developing instance
                --lisp             enable Lots of Irritating Superfluous Parentheses
                --new-proxy        don't use the saved proxy
                --new-session      start a new session with saved URL and key
                --no-config        start without loading/saving the config
                --no-patch-help    don't patch help()
                --reset-config     reset the whole config
                --timing           print the time it took to execute each command

                --help or -h       show this help message
            """
        )
    for arg in sys.argv[1:]:
        if arg not in {
            "--dev",
            "--lisp",
            "--new-proxy",
            "--new-session",
            "--no-config",
            "--no-patch-help",
            "--reset-config",
            "--timing",
            "--help",
            "-h",
        }:
            print(f"Unknown argument: {arg}")
            sys.exit(64 + 4 + 1)
    file: io.IOBase
    url: None | str = None
    key: None | str = None
    session: None | str = None
    proxy_type: None | int | EllipsisType = None
    proxy_addr: None | str = None
    proxy_port: None | int = None
    proxy_rdns: None | bool = True
    proxy_username: None | str = None
    proxy_password: None | str = None
    config_pickle = os.path.join(
        os.path.expanduser(os.getenv("XDG_CONFIG_HOME") or "~/.config"),
        "an-backdoor-client/"
        + ("dev-" if "--dev" in sys.argv else "")
        + "session.pickle",
    )
    if "--reset-config" in sys.argv and os.path.exists(config_pickle):
        os.remove(config_pickle)
    if "--no-config" not in sys.argv:
        try:
            with open(config_pickle, "rb") as file:
                config = pickle.load(file)  # nosec: B301
        except FileNotFoundError:
            pass
        else:
            url = config.get("url")
            key = config.get("key")
            session = config.get("session")
            proxy_type = config.get("proxy_type")
            proxy_addr = config.get("proxy_addr")
            proxy_port = config.get("proxy_port")
            proxy_rdns = config.get("proxy_rdns")
            proxy_username = config.get("proxy_username")
            proxy_password = config.get("proxy_password")
            if "--new-session" in sys.argv:
                print(f"Using URL {url}")
            else:
                print(f"Using URL {url} with existing session {session}")
    while not url:  # pylint: disable=while-used
        url = input("URL: ").strip().rstrip("/")
        if not url:
            print("No URL given!")
        elif "://" not in url:
            if not url.startswith("//"):
                url = "//" + url
            if re.match(
                r"^(\/\/)(localhost|127\.0\.0\.1|\[::1\])(\:\d+)?(/\S*)?$", url
            ):
                url = "http:" + url
            else:
                url = "https:" + url
            print(f"Using URL {url}")

    while not key:  # pylint: disable=while-used
        key = input("Key: ").strip()
        if not key:
            print("No key given!")

    if proxy_type is None or "--new-proxy" in sys.argv:
        proxy_url_str = input("Proxy (leave empty for none): ").strip()
        if proxy_url_str:
            if "://" not in proxy_url_str:
                if not proxy_url_str.startswith("//"):
                    proxy_url_str = "//" + proxy_url_str
                proxy_url_str = "socks5:" + proxy_url_str
            proxy_url = urlsplit(proxy_url_str)
            if proxy_url.hostname:
                proxy_type = int(socks.PROXY_TYPES[proxy_url.scheme.upper()])
                proxy_addr = proxy_url.hostname
                proxy_port = proxy_url.port
                proxy_rdns = True
                proxy_username = proxy_url.username or None
                proxy_password = proxy_url.password or None
            else:
                print("Invalid proxy URL!")
        else:
            print("No proxy given!")
    if isinstance(proxy_type, EllipsisType):
        proxy_type = None
    if proxy_type is not None:
        print(
            f"Using {socks.PRINTABLE_PROXY_TYPES[proxy_type]} proxy "
            f"{proxy_addr}{f':{proxy_port}' if proxy_port else ''}"
            + (f" with username {proxy_username}" if proxy_username else "")
        )
    else:
        print("Using no proxy (use --new-proxy to be able to set one)")

    if not session or "--new-session" in sys.argv:
        session = input("Session (enter nothing for random session): ")
        if not session:
            session = str(uuid.uuid4())
        print(f"Using session {session}")

    if "--no-config" not in sys.argv:
        os.makedirs(os.path.dirname(config_pickle), exist_ok=True)
        with open(config_pickle, "wb") as file:
            pickle.dump(
                {
                    "url": url,
                    "key": key,
                    "session": session,
                    "proxy_type": proxy_type or ...,  # not None (None == ask)
                    "proxy_addr": proxy_addr,
                    "proxy_port": proxy_port,
                    "proxy_rdns": proxy_rdns,
                    "proxy_username": proxy_username,
                    "proxy_password": proxy_password,
                },
                file,
                5,
            )

    proxy: None | Proxy

    if proxy_type and proxy_addr:
        proxy = {"type": proxy_type, "addr": proxy_addr}
        if proxy_port:
            proxy["port"] = proxy_port
        if proxy_rdns:
            proxy["rdns"] = proxy_rdns
        if proxy_username and proxy_password:
            proxy["username"] = proxy_username
            proxy["password"] = proxy_password
    else:
        proxy = None

    def send_to_remote(code: str, *, mode: str) -> Any:
        """Send code to the remote backdoor and return the unpickled body."""
        return send(
            url,
            key,
            code,
            mode,
            session,
            proxy,
        )[2]

    if "--no-patch-help" not in sys.argv:
        body = send_to_remote(
            # fmt: off
            "class _HelpHelper_92005ecf3788faea8346a7919fba0232188561ab:\n"
            "  def __call__(self, *args, **kwargs):\n"
            "    import io\n"
            "    import pydoc\n"
            "    helper_output = io.StringIO()\n"
            "    pydoc.Helper(io.StringIO(), helper_output)(*args, **kwargs)\n"
            "    return 'PagerTuple', helper_output.getvalue()\n"
           f"  __str__ = __repr__ = lambda _:{repr(help)!r}\n"  # noqa: E131
            "help = _HelpHelper_92005ecf3788faea8346a7919fba0232188561ab()\n"
            "del _HelpHelper_92005ecf3788faea8346a7919fba0232188561ab",
            # fmt: on
            mode="exec",
        )
        if not (isinstance(body, dict) and body["success"]):
            print("\033[91mPatching help() failed!\033[0m")

    if "--lisp" in sys.argv:
        if not hy:
            sys.exit("\033[91mHy is not installed!\033[0m")
        body = send_to_remote("__import__('hy')", mode="exec")
        if not (isinstance(body, dict) and body["success"]):
            print("\033[91mInjecting Hy builtins failed!\033[0m")

    body = send_to_remote(
        "import sys\nprint('Python', sys.version, 'on', sys.platform)",
        mode="exec",
    )
    if isinstance(body, dict) and body["success"] and body["output"]:
        print(f"\033[92mConnection to {url} was successful.\033[0m")
        print(body["output"].strip())
    else:
        print("\033[91mGetting remote information failed.\033[0m")
    print(
        'Type "copyright", "credits" or '
        'use the "help" function for more information.'
    )

    # pylint: disable=import-outside-toplevel, import-error, useless-suppression
    from pyrepl.python_reader import (  # type: ignore[import, unused-ignore]
        ReaderConsole,
    )
    from pyrepl.python_reader import main as _main

    def _run_and_print(  # type: ignore[no-any-unimported]
        self: ReaderConsole,
        code: str,
    ) -> None:
        # pylint: disable=unused-argument
        try:
            run_and_print(
                url,
                key,
                shellify(code),
                "--lisp" in sys.argv,
                session,
                proxy,
                "--timing" in sys.argv,
            )
        except Exception:  # pylint: disable=broad-except
            print(
                "\033[91mAn unexpected error occurred. "
                "Please contact a developer.\033[0m"
            )
            traceback.print_exc()

    # pylint: disable=redefined-variable-type

    # patch the reader console to use our function
    rc_execute = ReaderConsole.execute
    ReaderConsole.execute = _run_and_print

    # run the reader
    with contextlib.suppress(EOFError):
        _main(print_banner=False, clear_main=False)

    # restore the original method
    ReaderConsole.execute = rc_execute

    return 0


if __name__ == "__main__":
    sys.exit(main())



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Logging stuff used by the website."""

from __future__ import annotations

import asyncio
import logging
import os
from asyncio import AbstractEventLoop
from collections.abc import Awaitable
from concurrent.futures import Future
from datetime import datetime, tzinfo
from logging import LogRecord
from typing import Any

import orjson as json
from tornado.httpclient import AsyncHTTPClient

from .. import DIR as ROOT_DIR


class AsyncHandler(logging.Handler):
    """A logging handler that can handle log records asynchronously."""

    futures: set[Future[Any]]
    loop: AbstractEventLoop

    def __init__(
        self,
        level: int | str = logging.NOTSET,
        *,
        loop: AbstractEventLoop,
    ):
        """Initialize the handler."""
        super().__init__(level=level)
        self.futures = set()
        self.loop = loop

    def callback(self, future: Future[Any]) -> None:
        """Remove the reference to the future from the handler."""
        self.acquire()
        try:
            self.futures.discard(future)
        finally:
            self.release()

    def emit(  # type: ignore[override]
        self, record: LogRecord
    ) -> None | Awaitable[Any]:
        """
        Do whatever it takes to actually log the specified logging record.

        This version is intended to be implemented by subclasses and so
        raises a NotImplementedError.
        """
        raise NotImplementedError(
            "emit must be implemented by AsyncHandler subclasses"
        )

    def handle(  # type: ignore[override]
        self, record: LogRecord
    ) -> bool | LogRecord:
        """Handle incoming log records."""
        rv = self.filter(record)
        if isinstance(rv, LogRecord):
            record = rv
        if rv and not self.loop.is_closed():
            self.acquire()
            try:
                if awaitable := self.emit(record):
                    future = asyncio.run_coroutine_threadsafe(
                        awaitable, self.loop
                    )
                    self.futures.add(future)
                    future.add_done_callback(self.callback)
            finally:
                self.release()
        return rv


class DatetimeFormatter(logging.Formatter):
    """A logging formatter that formats the time using datetime."""

    timezone: None | tzinfo = None

    def formatTime(  # noqa: N802
        self, record: LogRecord, datefmt: None | str = None
    ) -> str:
        """Return the creation time of the LogRecord as formatted text."""
        spam = datetime.fromtimestamp(record.created).astimezone(self.timezone)
        if datefmt:
            return spam.strftime(datefmt)
        return spam.isoformat()


class WebhookFormatter(DatetimeFormatter):
    """A logging formatter optimized for logging to a webhook."""

    escape_message = False

    def format(self, record: LogRecord) -> str:
        """Format the specified record as text."""
        record.message = record.getMessage()
        if self.escape_message:
            record.message = json.dumps(record.message).decode("UTF-8")[1:-1]
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)
        return self.formatMessage(record)


class WebhookHandler(AsyncHandler):
    """A logging handler that sends logs to a webhook."""

    url: str
    content_type: str

    def __init__(
        self,
        level: int | str = logging.NOTSET,
        *,
        loop: AbstractEventLoop,
        url: str,
        content_type: str,
    ):
        """Initialize the handler."""
        super().__init__(level=level, loop=loop)
        self.url = url
        self.content_type = content_type

    async def emit(self, record: LogRecord) -> None:  # type: ignore[override]
        """Send the request to the webhook."""
        # pylint: disable=invalid-overridden-method
        try:
            message = self.format(record)
            await AsyncHTTPClient().fetch(
                self.url,
                method="POST",
                headers={"Content-Type": self.content_type},
                body=message.strip(),
                ca_certs=os.path.join(ROOT_DIR, "ca-bundle.crt"),
            )
        except Exception:  # pylint: disable=broad-except
            self.handleError(record)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the request handlers of an-website."""

from __future__ import annotations

from http.cookies import SimpleCookie

import orjson as json

from . import (  # noqa: F401  # pylint: disable=unused-import
    FetchCallable,
    app,
    assert_url_query,
    assert_valid_response,
    fetch,
)


async def test_setting_stuff_without_cookies(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Test changing settings with requests without saving to cookie."""
    for request_body in (
        json.dumps(
            {
                "bumpscosity": "0",
                "theme": "pink",
                "no_3rd_party": "sure",
                "dynload": "sure",
                "openmoji": "img",
            }
        ),
        json.dumps(
            {
                "bumpscosity": "0",
                "theme": "pink",
                "no_3rd_party": "sure",
                "dynload": "sure",
                "openmoji": "img",
                "save_in_cookie": "nope",
            }
        ),
        "theme=pink&no_3rd_party=sure&dynload=sure&openmoji=img&bumpscosity=0",
        (
            "theme=pink&no_3rd_party=s&dynload=s&save_in_cookie=n&openmoji=i"
            "&bumpscosity=0"
        ),
        "",
    ):
        response = await fetch(
            "/einstellungen"
            if request_body
            else (
                "/einstellungen?theme=pink&no_3rd_party=s&dynload=s&openmoji=i"
                "&bumpscosity=0"
            ),
            method="POST",
            headers={"Content-Type": "application/json"}
            if isinstance(request_body, bytes)
            else None,
            body=request_body,
            follow_redirects=True,
        )
        assert_valid_response(response, content_type="text/html;charset=utf-8")
        assert_url_query(
            response.effective_url,
            theme="pink",
            no_3rd_party="sure",
            dynload="sure",
            openmoji="img",
            save_in_cookie="nope",
            bumpscosity="0",
        )


def parse_cookie(cookie: str) -> SimpleCookie:
    """Parse a cookie string to a SimpleCookie."""
    simple_cookie = SimpleCookie()
    simple_cookie.load(cookie)
    return simple_cookie


async def test_setting_stuff_and_saving_to_cookies(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Test changing settings with requests with saving to cookie."""
    for request_body in (
        json.dumps(
            {
                "theme": "pink",
                "no_3rd_party": "sure",
                "dynload": "sure",
                "save_in_cookie": "sure",
                "openmoji": "img",
                "bumpscosity": "0",
            }
        ),
        "theme=pink&no_3rd_party=s&dynload=s&save_in_cookie=s&openmoji=i&bumpscosity=0",
        "",
    ):
        response = await fetch(
            "/einstellungen"
            + (
                "?theme=pink&no_3rd_party=s&dynload=s&openmoji=i&save_in_cookie=s&bumpscosity=0"
                if not request_body
                else ""
            ),
            method="POST",
            headers={"Content-Type": "application/json"}
            if isinstance(request_body, bytes)
            else None,
            body=request_body,
        )
        if request_body:
            assert_valid_response(
                response, content_type="text/html;charset=utf-8"
            )
        else:
            assert response.headers["Location"].endswith("/einstellungen")

        for cookie_header in response.headers.get_list("Set-Cookie"):
            cookie = parse_cookie(cookie_header)
            assert len(cookie) == 1
            morsel = cookie[tuple(cookie)[0]]
            assert morsel["expires"]
            assert morsel["path"] == "/"
            assert morsel["samesite"] == "Strict"
            assert morsel.key in {
                "advanced_settings",
                "bumpscosity",
                "compat",
                "dynload",
                "effects",
                "no_3rd_party",
                "openmoji",
                "theme",
            }
            if morsel.key == "theme":
                assert morsel.value == "pink"
            elif morsel.key == "openmoji":
                assert morsel.value == "img"
            elif morsel.key == "bumpscosity":
                assert morsel.value == "0"
            elif morsel.key in {"advanced_settings", "compat"}:
                assert morsel.value == "nope"
            else:
                assert morsel.value == "sure"



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# pylint: disable=protected-access

"""Patches that improve everything."""

from __future__ import annotations

import http.client
import json as stdlib_json  # pylint: disable=preferred-module
import os
import sys
from collections.abc import Callable
from configparser import RawConfigParser
from contextlib import suppress
from importlib import import_module
from os.path import abspath, dirname
from pathlib import Path
from threading import Thread
from types import MethodType
from typing import Any, Final
from urllib.parse import urlsplit

import certifi
import defusedxml  # type: ignore[import-untyped]
import jsonpickle  # type: ignore[import-untyped]
import orjson
import pycurl
import tornado.httputil
import yaml
from emoji import EMOJI_DATA
from emoji import unicode_codes as euc
from pillow_jxl import JpegXLImagePlugin  # noqa: F401
from setproctitle import setthreadtitle
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.httputil import HTTPFile, HTTPHeaders, HTTPServerRequest
from tornado.log import gen_log
from tornado.web import GZipContentEncoding, RedirectHandler, RequestHandler

from .. import DIR as ROOT_DIR
from .. import MEDIA_TYPES
from . import braille, json  # noqa: F401  # pylint: disable=reimported

DIR: Final = abspath(dirname(__file__))


def apply() -> None:
    """Improve."""
    patch_asyncio()
    patch_certifi()
    patch_configparser()
    patch_emoji()
    patch_http()
    patch_json()
    patch_jsonpickle()
    patch_threading()
    patch_xml()

    patch_tornado_418()
    patch_tornado_arguments()
    patch_tornado_gzip()
    patch_tornado_httpclient()
    patch_tornado_logs()
    patch_tornado_redirect()


def patch_asyncio() -> None:
    """Make stuff faster."""
    if "DISABLE_UVLOOP" not in os.environ:
        with suppress(ModuleNotFoundError):
            import_module("uvloop").install()


def patch_certifi() -> None:
    """Make everything use our CA bundle."""
    certifi.where = lambda: os.path.join(ROOT_DIR, "ca-bundle.crt")
    certifi.contents = lambda: Path(certifi.where()).read_text("ASCII")


def patch_configparser() -> None:
    """Make configparser funky."""
    RawConfigParser.BOOLEAN_STATES.update(  # type: ignore[attr-defined]
        {
            "sure": True,
            "nope": False,
            "accept": True,
            "reject": False,
            "enabled": True,
            "disabled": False,
        }
    )


def patch_emoji() -> None:
    """Add cool new emoji."""
    EMOJI_DATA["🐱\u200D💻"] = {
        "de": ":hacker_katze:",
        "en": ":hacker_cat:",
        "status": 2,
        "E": 1,
    }
    for de_name, en_name, rect in (
        ("rot", "red", "🟥"),
        ("blau", "blue", "🟦"),
        ("orang", "orange", "🟧"),
        ("gelb", "yellow", "🟨"),
        ("grün", "green", "🟩"),
        ("lilan", "purple", "🟪"),
        ("braun", "brown", "🟫"),
    ):
        EMOJI_DATA[f"🫙\u200D{rect}"] = {
            "de": f":{de_name}es_glas:",
            "en": f":{en_name}_jar:",
            "status": 2,
            "E": 14,
        }
        EMOJI_DATA[f"🏳\uFE0F\u200D{rect}"] = {
            "de": f":{de_name}e_flagge:",
            "en": f":{en_name}_flag:",
            "status": 2,
            "E": 11,
        }
        EMOJI_DATA[f"\u2691\uFE0F\u200D{rect}"] = {
            "de": f":tief{de_name}e_flagge:",
            "en": f":deep_{en_name}_flag:",
            "status": 2,
            "E": 11,
        }
    for lang in euc.LANGUAGES:
        euc._EMOJI_UNICODE[lang] = None  # type: ignore[attr-defined]
    euc._ALIASES_UNICODE.clear()  # type: ignore[attr-defined]


def patch_http() -> None:
    """Add response code 420."""
    http.client.responses[420] = "Enhance Your Calm"


def patch_json() -> None:
    """Replace json with orjson."""
    if getattr(stdlib_json, "_omegajson", False) or sys.version_info < (3, 12):
        return
    stdlib_json.dumps = json.dumps
    stdlib_json.dump = json.dump  # type: ignore[assignment]
    stdlib_json.loads = json.loads  # type: ignore[assignment]
    stdlib_json.load = json.load


def patch_jsonpickle() -> None:
    """Make jsonpickle return bytes."""
    jsonpickle.load_backend("orjson")
    jsonpickle.set_preferred_backend("orjson")
    jsonpickle.enable_fallthrough(False)


def patch_threading() -> None:
    """Set thread names."""
    _bootstrap = Thread._bootstrap  # type: ignore[attr-defined]

    def bootstrap(self: Thread) -> None:
        with suppress(Exception):
            setthreadtitle(self.name)
        _bootstrap(self)

    Thread._bootstrap = bootstrap  # type: ignore[attr-defined]


def patch_tornado_418() -> None:
    """Add support for RFC 7168."""
    RequestHandler.SUPPORTED_METHODS += (  # type: ignore[assignment]
        "PROPFIND",
        "BREW",
        "WHEN",
    )
    _ = RequestHandler._unimplemented_method
    RequestHandler.propfind = _  # type: ignore[attr-defined]
    RequestHandler.brew = _  # type: ignore[attr-defined]
    RequestHandler.when = _  # type: ignore[attr-defined]


def patch_tornado_arguments() -> None:  # noqa: C901
    """Improve argument parsing."""
    # pylint: disable=too-complex

    def ensure_bytes(value: Any) -> bytes:
        """Return the value as bytes."""
        if isinstance(value, bool):
            return b"true" if value else b"false"
        if isinstance(value, bytes):
            return value
        return str(value).encode("UTF-8")

    def parse_body_arguments(
        content_type: str,
        body: bytes,
        arguments: dict[str, list[bytes]],
        files: dict[str, list[HTTPFile]],
        headers: None | HTTPHeaders = None,
        *,
        _: Callable[..., None] = tornado.httputil.parse_body_arguments,
    ) -> None:
        # pylint: disable=too-many-branches
        if content_type.startswith("application/json"):
            if headers and "Content-Encoding" in headers:
                gen_log.warning(
                    "Unsupported Content-Encoding: %s",
                    headers["Content-Encoding"],
                )
                return
            try:
                spam = orjson.loads(body)
            except Exception as exc:  # pylint: disable=broad-except
                gen_log.warning("Invalid JSON body: %s", exc)
            else:
                for key, value in spam.items():
                    if value is not None:
                        arguments.setdefault(key, []).append(
                            ensure_bytes(value)
                        )
        elif content_type.startswith("application/yaml"):
            if headers and "Content-Encoding" in headers:
                gen_log.warning(
                    "Unsupported Content-Encoding: %s",
                    headers["Content-Encoding"],
                )
                return
            try:
                spam = yaml.safe_load(body)
            except Exception as exc:  # pylint: disable=broad-except
                gen_log.warning("Invalid YAML body: %s", exc)
            else:
                for key, value in spam.items():
                    if value is not None:
                        arguments.setdefault(key, []).append(
                            ensure_bytes(value)
                        )
        else:
            _(content_type, body, arguments, files, headers)

    parse_body_arguments.__doc__ = tornado.httputil.parse_body_arguments.__doc__

    tornado.httputil.parse_body_arguments = parse_body_arguments


def patch_tornado_gzip() -> None:
    """Use gzip for more content types."""
    GZipContentEncoding.CONTENT_TYPES = {
        type for type, data in MEDIA_TYPES.items() if data.get("compressible")
    }


def patch_tornado_httpclient() -> None:
    """Make requests quick."""
    BACON = 0x75800  # noqa: N806  # pylint: disable=invalid-name
    EGGS = 1 << 25  # noqa: N806  # pylint: disable=invalid-name

    AsyncHTTPClient.configure("tornado.curl_httpclient.CurlAsyncHTTPClient")

    def prepare_curl_callback(self: HTTPRequest, curl: pycurl.Curl) -> None:
        # pylint: disable=c-extension-no-member, useless-suppression
        if urlsplit(self.url).scheme == "https":  # noqa: SIM102
            if (ver := pycurl.version_info())[2] >= BACON and ver[4] & EGGS:
                curl.setopt(pycurl.HTTP_VERSION, pycurl.CURL_HTTP_VERSION_3)

    original_request_init = HTTPRequest.__init__

    def request_init(self: HTTPRequest, *args: Any, **kwargs: Any) -> None:
        if len(args) < 18:  # there are too many positional arguments here
            prepare_curl_method = MethodType(prepare_curl_callback, self)
            kwargs.setdefault("prepare_curl_callback", prepare_curl_method)
        original_request_init(self, *args, **kwargs)

    request_init.__doc__ = HTTPRequest.__init__.__doc__

    HTTPRequest.__init__ = request_init  # type: ignore[method-assign]


def patch_tornado_logs() -> None:
    """Anonymize Tornado logs."""
    # pylint: disable=import-outside-toplevel
    from ..utils.utils import SUS_PATHS, anonymize_ip

    RequestHandler._request_summary = (  # type: ignore[method-assign]
        lambda self: "%s %s (%s)"  # pylint: disable=consider-using-f-string
        % (
            self.request.method,
            self.request.uri,
            self.request.remote_ip
            if self.request.path == "/robots.txt"
            or self.request.path.lower() in SUS_PATHS
            else anonymize_ip(self.request.remote_ip, ignore_invalid=True),
        )
    )

    HTTPServerRequest.__repr__ = (  # type: ignore[method-assign]
        lambda self: "%s(%s)"  # pylint: disable=consider-using-f-string
        % (
            self.__class__.__name__,
            ", ".join(
                [
                    "%s=%r"  # pylint: disable=consider-using-f-string
                    % (
                        n,
                        getattr(self, n),
                    )
                    for n in ("protocol", "host", "method", "uri", "version")
                ]
            ),
        )
    )


def patch_tornado_redirect() -> None:
    """Use modern redirect codes and support HEAD requests."""

    def redirect(
        self: RequestHandler,
        url: str,
        permanent: bool = False,
        status: None | int = None,
    ) -> None:
        if self._headers_written:
            # pylint: disable=broad-exception-raised
            raise Exception("Cannot redirect after headers have been written")
        if status is None:
            status = 308 if permanent else 307
        else:
            assert isinstance(status, int) and 300 <= status <= 399  # type: ignore[redundant-expr]  # noqa: B950
        self.set_status(status)
        self.set_header("Location", url)
        self.finish()  # type: ignore[unused-awaitable]

    if RequestHandler.redirect.__doc__:
        # fmt: off
        redirect.__doc__ = (
            RequestHandler.redirect.__doc__
            .replace("301", "308")
            .replace("302", "307")
        )
        # fmt: on

    RequestHandler.redirect = redirect  # type: ignore[method-assign]

    RedirectHandler.head = RedirectHandler.get


def patch_xml() -> None:
    """Make XML safer."""
    defusedxml.defuse_stdlib()
    defusedxml.xmlrpc.monkey_patch()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
A page with all the JavaScript files and their licenses.

This is used for LibreJS to make sure the extension knows the licenses.
This isn't important, as the JS files should contain the licenses themselves.

See: https://www.gnu.org/software/librejs/free-your-javascript.html#step3
and https://www.gnu.org/licenses/javascript-labels.html
"""

from __future__ import annotations

import logging
import os
from collections.abc import Mapping
from functools import cache
from pathlib import Path
from typing import Final

from .. import STATIC_DIR
from ..utils.request_handler import HTMLRequestHandler
from ..utils.utils import ModuleInfo

LOGGER: Final = logging.getLogger(__name__)


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=((r"/js-lizenzen", JSLicenses),),
        name="JavaScript-Lizenzen",
        description=(
            "Informationen über die Lizenzen der JavaScript-Dateien "
            "auf dieser Seite"
        ),
        path="/js-lizenzen",
        aliases=("/js-licenses",),
        keywords=("JavaScript", "License", "Lizenz"),
    )


LICENSES: Final[Mapping[str, str]] = {
    "magnet:?xt=urn:btih:0b31508aeb0634b347b8270c7bee4d411b5d4109&dn=agpl-3.0.txt": (
        "https://www.gnu.org/licenses/agpl-3.0.html"
    ),
    "magnet:?xt=urn:btih:d3d9a9a6595521f9666a5e94cc830dab83b65699&dn=expat.txt": (
        "https://www.jclark.com/xml/copying.txt"
    ),
}


@cache
def get_js_filenames_and_licenses() -> list[tuple[str, str, str]]:
    """
    Get the names of the JS files in this project.

    Returns a list of tuples with filename, license and license URL.
    """
    js_files_dir = os.path.join(STATIC_DIR, "js")
    licenses_list: list[tuple[str, str, str]] = []
    for path in Path(js_files_dir).rglob("*.js"):
        if not path.is_file():
            continue
        filename = str(path.relative_to(js_files_dir))
        with path.open(encoding="UTF-8") as file:
            license_line = file.readline().strip().removeprefix('"use strict";')
        if not license_line.startswith("// @license "):
            LOGGER.warning("%s has no license comment", filename)
            # TODO: exit in dev mode if this fails
            continue
        magnet, name = (
            license_line.removeprefix("// @license").strip().split(" ")
        )
        licenses_list.append((filename, name.strip(), LICENSES[magnet.strip()]))
    return licenses_list


class JSLicenses(HTMLRequestHandler):
    """The request handler for the JS licenses page."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the JS licenses page."""
        if head:
            return
        await self.render(
            "pages/js_licenses.html", js_files=get_js_filenames_and_licenses()
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A page with wrong quotes."""

from __future__ import annotations

import abc
import asyncio
import contextlib
import logging
import os
import random
import sys
import time
from collections.abc import Callable, Iterable, Mapping, Sequence
from dataclasses import dataclass
from datetime import date, datetime, timezone
from multiprocessing import Value
from typing import Any, Final, Literal, cast
from urllib.parse import urlencode

import elasticapm  # type: ignore[import-untyped]
import orjson as json
from redis.asyncio import Redis
from tornado.httpclient import AsyncHTTPClient
from tornado.web import Application, HTTPError
from UltraDict import UltraDict  # type: ignore[import-untyped]

from .. import DIR as ROOT_DIR
from .. import (
    EVENT_REDIS,
    EVENT_SHUTDOWN,
    NAME,
    ORJSON_OPTIONS,
    pytest_is_running,
)
from ..utils.request_handler import HTMLRequestHandler
from ..utils.utils import Permission, emojify, ratelimit

DIR: Final = os.path.dirname(__file__)

LOGGER: Final = logging.getLogger(__name__)

API_URL: Final[str] = "https://zitate.prapsschnalinen.de/api"

QUOTES_CACHE: Final[dict[int, Quote]] = UltraDict(buffer_size=1024**2)
AUTHORS_CACHE: Final[dict[int, Author]] = UltraDict(buffer_size=1024**2)
WRONG_QUOTES_CACHE: Final[dict[tuple[int, int], WrongQuote]] = UltraDict(
    buffer_size=1024**2
)

MAX_QUOTES_ID = Value("Q", 0)
MAX_AUTHORS_ID = Value("Q", 0)


@dataclass(init=False, slots=True)
class QuotesObjBase(abc.ABC):
    """An object with an id."""

    id: int

    @abc.abstractmethod
    async def fetch_new_data(self) -> QuotesObjBase:
        """Fetch new data from the API."""
        raise NotImplementedError

    # pylint: disable=unused-argument
    def get_id_as_str(self, minify: bool = False) -> str:
        """Get the id of the object as a string."""
        return str(self.id)

    @abc.abstractmethod
    def get_path(self) -> str:
        """Return the path to the Object."""
        raise NotImplementedError


@dataclass(slots=True)
class Author(QuotesObjBase):
    """The author object with a name."""

    name: str
    # tuple(url_to_info, info_str, creation_date)
    info: None | tuple[str, None | str, date]

    def __str__(self) -> str:
        """Return the name of the author."""
        return (
            emojify(self.name.strip())
            if (now := datetime.now(timezone.utc)).day == 1 and now.month == 4
            else self.name.strip()
        )

    async def fetch_new_data(self) -> Author:
        """Fetch new data from the API."""
        return parse_author(await make_api_request(f"authors/{self.id}"))

    def get_path(self) -> str:
        """Return the path to the author info."""
        return f"/zitate/info/a/{self.id}"

    def to_json(self) -> dict[str, Any]:
        """Get the author as JSON."""
        return {
            "id": self.id,
            "name": str(self),
            "path": self.get_path(),
            "info": {
                "source": self.info[0],
                "text": self.info[1],
                "date": self.info[2].isoformat(),
            }
            if self.info
            else None,
        }

    def update_name(self, name: str) -> None:
        """Update author data with another author."""
        name = fix_author_name(name)
        if self.name != name:
            # name changed -> info should change too
            self.info = None
            self.name = name


@dataclass(slots=True)
class Quote(QuotesObjBase):
    """The quote object with a quote text and an author."""

    quote: str
    author: Author

    def __str__(self) -> str:
        """Return the content of the quote."""
        return (
            emojify(self.quote.strip())
            if (now := datetime.now(timezone.utc)).day == 1 and now.month == 4
            else self.quote.strip()
        )

    async def fetch_new_data(self) -> Quote:
        """Fetch new data from the API."""
        return parse_quote(await make_api_request(f"quotes/{self.id}"), self)

    def get_path(self) -> str:
        """Return the path to the quote info."""
        return f"/zitate/info/z/{self.id}"

    def to_json(self) -> dict[str, Any]:
        """Get the quote as JSON."""
        return {
            "id": self.id,
            "quote": str(self),
            "author": self.author.to_json(),
            "path": self.get_path(),
        }

    def update_quote(
        self, quote: str, author_id: int, author_name: str
    ) -> None:
        """Update quote data with new data."""
        self.quote = fix_quote_str(quote)
        if self.author.id == author_id:
            self.author.update_name(author_name)
            return
        self.author = get_author_updated_with(author_id, author_name)


@dataclass(slots=True)
class WrongQuote(QuotesObjBase):
    """The wrong quote object with a quote, an author and a rating."""

    quote: Quote
    author: Author
    rating: int

    def __str__(self) -> str:
        r"""
        Return the wrong quote.

        like: '»quote« - author'.
        """
        return f"»{self.quote}« - {self.author}"

    async def fetch_new_data(self) -> WrongQuote:
        """Fetch new data from the API."""
        if self.id == -1:
            api_data = await make_api_request(
                "wrongquotes",
                {
                    "quote": str(self.quote.id),
                    "simulate": "true",
                    "author": str(self.author.id),
                },
            )
            if api_data:
                api_data = api_data[0]
        else:
            api_data = await make_api_request(f"wrongquotes/{self.id}")
        if not api_data:
            return self
        return parse_wrong_quote(api_data, self)

    def get_id(self) -> tuple[int, int]:
        """
        Get the id of the quote and the author in a tuple.

        :return tuple(quote_id, author_id)
        """
        return self.quote.id, self.author.id

    def get_id_as_str(self, minify: bool = False) -> str:
        """
        Get the id of the wrong quote as a string.

        Format: quote_id-author_id
        """
        if minify and self.id != -1:
            return str(self.id)
        return f"{self.quote.id}-{self.author.id}"

    def get_path(self) -> str:
        """Return the path to the wrong quote."""
        return f"/zitate/{self.get_id_as_str()}"

    def to_json(self) -> dict[str, Any]:
        """Get the wrong quote as JSON."""
        return {
            "id": self.get_id_as_str(),
            "quote": self.quote.to_json(),
            "author": self.author.to_json(),
            "rating": self.rating,
            "path": self.get_path(),
        }

    async def vote(
        # pylint: disable=unused-argument
        self,
        vote: Literal[-1, 1],
        lazy: bool = False,
    ) -> WrongQuote:
        """Vote for the wrong quote."""
        if self.id == -1:
            raise ValueError("Can't vote for a not existing quote.")
        # if lazy:  # simulate the vote and do the actual voting later
        #     self.rating += vote
        #     asyncio.get_running_loop().call_soon_threadsafe(
        #         self.vote,
        #         vote,
        #     )
        #     return self
        # do the voting
        return parse_wrong_quote(
            await make_api_request(
                f"wrongquotes/{self.id}",
                method="POST",
                body={"vote": str(vote)},
            ),
            self,
        )


def get_wrong_quotes(
    filter_fun: None | Callable[[WrongQuote], bool] = None,
    *,
    sort: bool = False,  # sorted by rating
    filter_real_quotes: bool = True,
    shuffle: bool = False,
) -> Sequence[WrongQuote]:
    """Get cached wrong quotes."""
    if shuffle and sort:
        raise ValueError("Sort and shuffle can't be both true.")
    wqs: Iterable[WrongQuote] = WRONG_QUOTES_CACHE.values()
    if filter_fun is not None:
        # pylint: disable=bad-builtin
        wqs = filter(filter_fun, wqs)
    if filter_real_quotes:
        # pylint: disable=bad-builtin
        wqs = filter(lambda wq: wq.quote.author.id != wq.author.id, wqs)
    if not (shuffle or sort):
        return tuple(wqs)
    wqs_list = list(wqs)  # shuffle or sort is True
    if shuffle:
        random.shuffle(wqs_list)
    if sort:
        wqs_list.sort(key=lambda wq: wq.rating, reverse=True)
    return wqs_list


def get_quotes(
    filter_fun: None | Callable[[Quote], bool] = None,
    shuffle: bool = False,
) -> list[Quote]:
    """Get cached quotes."""
    quotes: list[Quote] = list(QUOTES_CACHE.values())
    if filter_fun is not None:
        quotes = [_q for _q in quotes if filter_fun(_q)]
    if shuffle:
        random.shuffle(quotes)
    return quotes


def get_authors(
    filter_fun: None | Callable[[Author], bool] = None,
    shuffle: bool = False,
) -> list[Author]:
    """Get cached authors."""
    authors: list[Author] = list(AUTHORS_CACHE.values())
    if filter_fun is not None:
        authors = [_a for _a in authors if filter_fun(_a)]
    if shuffle:
        random.shuffle(authors)
    return list(authors)


async def make_api_request(
    endpoint: str,
    args: Mapping[str, str] | None = None,
    *,
    method: Literal["GET", "POST"] = "GET",
    body: None | Mapping[str, str] = None,
) -> Any:  # TODO: list[dict[str, Any]] | dict[str, Any]:
    """Make API request and return the result as dict."""
    if pytest_is_running():
        return None
    query = f"?{urlencode(args)}" if args else ""
    url = f"{API_URL}/{endpoint}{query}"
    body_str = urlencode(body) if body else body
    response = await AsyncHTTPClient().fetch(
        url,
        method=method,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        body=body_str,
        raise_error=False,
        ca_certs=os.path.join(ROOT_DIR, "ca-bundle.crt"),
    )
    if response.code != 200:
        LOGGER.warning(
            "%s request to %r with body=%r failed with code=%d and reason=%r",
            method,
            url,
            body_str,
            response.code,
            response.reason,
        )
        raise HTTPError(
            400
            if response.code == 500
            else (404 if response.code == 404 else 503),
            reason=(
                f"{API_URL}/{endpoint} returned: "
                f"{response.code} {response.reason}"
            ),
        )
    return json.loads(response.body)


def fix_author_name(name: str) -> str:
    """Fix common mistakes in authors."""
    if len(name) > 2 and name.startswith("(") and name.endswith(")"):
        # remove () from author name, that shouldn't be there
        name = name[1:-1]
    return name.strip()


def get_author_updated_with(author_id: int, name: str) -> Author:
    """Get the author with the given id and the name."""
    name = fix_author_name(name)
    if not name:
        name = "None"

    author = AUTHORS_CACHE.get(author_id)
    if author is None:  # author not in cache, create new one
        # pylint: disable=too-many-function-args
        author = Author(author_id, name, None)
        MAX_AUTHORS_ID.value = max(  # type: ignore[attr-defined]
            MAX_AUTHORS_ID.value, author_id  # type: ignore[attr-defined]
        )
    else:  # update to make sure cache is correct
        author.update_name(name)

    AUTHORS_CACHE[author.id] = author

    return author


def parse_author(json_data: Mapping[str, Any]) -> Author:
    """Parse an author from JSON data."""
    return get_author_updated_with(int(json_data["id"]), json_data["author"])


def fix_quote_str(quote_str: str) -> str:
    """Fix common mistakes in quotes."""
    if (
        len(quote_str) > 2
        and quote_str.startswith(('"', "„", "“"))
        and quote_str.endswith(('"', "“", "”"))
    ):
        # remove quotation marks from quote, that shouldn't be there
        quote_str = quote_str[1:-1]

    return quote_str.strip()


def parse_quote(
    json_data: Mapping[str, Any], quote: None | Quote = None
) -> Quote:
    """Parse a quote from JSON data."""
    quote_id = int(json_data["id"])
    author = parse_author(json_data["author"])  # update author
    quote_str = fix_quote_str(json_data["quote"])

    if quote is None:  # no quote supplied, try getting it from cache
        quote = QUOTES_CACHE.get(quote_id)
    if quote is None:  # new quote
        # pylint: disable=too-many-function-args
        quote = Quote(quote_id, quote_str, author)
        MAX_QUOTES_ID.value = max(  # type: ignore[attr-defined]
            MAX_QUOTES_ID.value, quote.id  # type: ignore[attr-defined]
        )
    else:  # quote was already saved
        quote.update_quote(quote_str, author.id, author.name)

    QUOTES_CACHE[quote.id] = quote

    return quote


def parse_wrong_quote(
    json_data: Mapping[str, Any], wrong_quote: None | WrongQuote = None
) -> WrongQuote:
    """Parse a wrong quote and update the cache."""
    quote = parse_quote(json_data["quote"])
    author = parse_author(json_data["author"])

    id_tuple = (quote.id, author.id)
    rating = json_data["rating"]
    wrong_quote_id = int(json_data.get("id") or -1)

    if wrong_quote is None:
        wrong_quote = WRONG_QUOTES_CACHE.get(id_tuple)
        if wrong_quote is None:
            wrong_quote = WrongQuote(  # pylint: disable=unexpected-keyword-arg
                id=wrong_quote_id,
                quote=quote,
                author=author,
                rating=rating,
            )

    # make sure the wrong quote is the correct one
    if (wrong_quote.quote.id, wrong_quote.author.id) != id_tuple:
        raise HTTPError(reason="ERROR: -41")

    wrong_quote.quote = quote
    wrong_quote.author = author

    # update the data of the wrong quote
    if wrong_quote.rating != rating:
        wrong_quote.rating = rating
    if wrong_quote.id != wrong_quote_id:
        wrong_quote.id = wrong_quote_id

    WRONG_QUOTES_CACHE[id_tuple] = wrong_quote

    return wrong_quote


async def parse_list_of_quote_data(
    json_list: str | Iterable[Mapping[str, Any]],
    parse_fun: Callable[[Mapping[str, Any]], QuotesObjBase],
) -> tuple[QuotesObjBase, ...]:
    """Parse a list of quote data."""
    if not json_list:
        return ()
    if isinstance(json_list, str):
        json_list = cast(list[dict[str, Any]], json.loads(json_list))
    return_list = []
    for json_data in json_list:
        _ = parse_fun(json_data)
        await asyncio.sleep(0)
        return_list.append(_)
    return tuple(return_list)


async def update_cache_periodically(app: Application) -> None:
    """Start updating the cache every hour."""
    # pylint: disable=too-complex
    with contextlib.suppress(asyncio.TimeoutError):
        await asyncio.wait_for(EVENT_REDIS.wait(), 5)
    redis: Redis[str] = cast("Redis[str]", app.settings.get("REDIS"))
    prefix: str = app.settings.get("REDIS_PREFIX", NAME).removesuffix("-dev")
    apm: None | elasticapm.Client  # type: ignore[no-any-unimported]
    if EVENT_REDIS.is_set():  # pylint: disable=too-many-nested-blocks
        await parse_list_of_quote_data(
            await redis.get(f"{prefix}:cached-quote-data:wrongquotes"),  # type: ignore[arg-type]  # noqa: B950
            parse_wrong_quote,
        )
        await parse_list_of_quote_data(
            await redis.get(f"{prefix}:cached-quote-data:quotes"),  # type: ignore[arg-type]  # noqa: B950
            parse_quote,
        )
        await parse_list_of_quote_data(
            await redis.get(f"{prefix}:cached-quote-data:authors"),  # type: ignore[arg-type]  # noqa: B950
            parse_author,
        )
        if QUOTES_CACHE and AUTHORS_CACHE and WRONG_QUOTES_CACHE:
            last_update = await redis.get(
                f"{prefix}:cached-quote-data:last-update"
            )
            if last_update:
                last_update_int = int(last_update)
                since_last_update = int(time.time()) - last_update_int
                if 0 <= since_last_update < 60 * 60:
                    # wait until the last update is at least one hour old
                    update_cache_in = 60 * 60 - since_last_update
                    if not sys.flags.dev_mode and update_cache_in > 60:
                        # if in production mode update wrong quotes just to be sure
                        try:
                            await update_cache(
                                app, update_quotes=False, update_authors=False
                            )
                        except Exception:  # pylint: disable=broad-except
                            LOGGER.exception("Updating quotes cache failed")
                            apm = app.settings.get("ELASTIC_APM", {}).get(
                                "CLIENT"
                            )
                            if apm:
                                apm.capture_exception()
                        else:
                            LOGGER.info("Updated quotes cache successfully")
                    LOGGER.info(
                        "Next update of quotes cache in %d seconds",
                        update_cache_in,
                    )
                    await asyncio.sleep(update_cache_in)

    # update the cache every hour
    failed = 0
    while not EVENT_SHUTDOWN.is_set():  # pylint: disable=while-used
        try:
            await update_cache(app)
        except Exception:  # pylint: disable=broad-except
            LOGGER.exception("Updating quotes cache failed")
            if apm := app.settings.get("ELASTIC_APM", {}).get("CLIENT"):
                apm.capture_exception()
            failed += 1
            await asyncio.sleep(pow(min(failed * 2, 60), 2))  # 4,16,...,60*60
        else:
            LOGGER.info("Updated quotes cache successfully")
            failed = 0
            await asyncio.sleep(60 * 60)


async def update_cache(
    app: Application,
    update_wrong_quotes: bool = True,
    update_quotes: bool = True,
    update_authors: bool = True,
) -> None:
    """Fill the cache with all data from the API."""
    LOGGER.info("Updating quotes cache")
    redis: Redis[str] = cast("Redis[str]", app.settings.get("REDIS"))
    prefix: str = app.settings.get("REDIS_PREFIX", NAME).removesuffix("-dev")
    redis_available = EVENT_REDIS.is_set()

    if update_wrong_quotes:
        await parse_list_of_quote_data(
            wq_data := await make_api_request("wrongquotes"),
            parse_wrong_quote,
        )
        if wq_data and redis_available:
            await redis.setex(
                f"{prefix}:cached-quote-data:wrongquotes",
                60 * 60 * 24 * 30,
                json.dumps(wq_data, option=ORJSON_OPTIONS),
            )

    if update_quotes:
        await parse_list_of_quote_data(
            quotes_data := await make_api_request("quotes"),
            parse_quote,
        )
        if quotes_data and redis_available:
            await redis.setex(
                f"{prefix}:cached-quote-data:quotes",
                60 * 60 * 24 * 30,
                json.dumps(quotes_data, option=ORJSON_OPTIONS),
            )

    if update_authors:
        await parse_list_of_quote_data(
            authors_data := await make_api_request("authors"),
            parse_author,
        )
        if authors_data and redis_available:
            await redis.setex(
                f"{prefix}:cached-quote-data:authors",
                60 * 60 * 24 * 30,
                json.dumps(authors_data, option=ORJSON_OPTIONS),
            )

    if (
        redis_available
        and update_wrong_quotes
        and update_quotes
        and update_authors
    ):
        await redis.setex(
            f"{prefix}:cached-quote-data:last-update",
            60 * 60 * 24 * 30,
            int(time.time()),
        )


async def get_author_by_id(author_id: int) -> Author:
    """Get an author by its id."""
    author = AUTHORS_CACHE.get(author_id)
    if author is not None:
        return author
    return parse_author(await make_api_request(f"authors/{author_id}"))


async def get_quote_by_id(quote_id: int) -> Quote:
    """Get a quote by its id."""
    quote = QUOTES_CACHE.get(quote_id)
    if quote is not None:
        return quote
    return parse_quote(await make_api_request(f"quotes/{quote_id}"))


async def get_wrong_quote(
    quote_id: int, author_id: int, use_cache: bool = True
) -> WrongQuote:
    """Get a wrong quote with a quote id and an author id."""
    wrong_quote = WRONG_QUOTES_CACHE.get((quote_id, author_id))
    if wrong_quote:
        if use_cache:
            return wrong_quote
        # do not use cache, so update the wrong quote data
        return await wrong_quote.fetch_new_data()
    # wrong quote not in cache
    if use_cache and quote_id in QUOTES_CACHE and author_id in AUTHORS_CACHE:
        # we don't need to request anything, as the wrong_quote probably has
        # no ratings just use the cached quote and author
        return WrongQuote(  # pylint: disable=too-many-function-args
            -1,
            QUOTES_CACHE[quote_id],
            AUTHORS_CACHE[author_id],
            0,
        )
    # request the wrong quote from the API
    result = await make_api_request(
        "wrongquotes",
        {
            "quote": str(quote_id),
            "simulate": "true",
            "author": str(author_id),
        },
    )
    if result:
        return parse_wrong_quote(result[0])

    raise HTTPError(404, reason="Falsches Zitat nicht gefunden")


async def get_rating_by_id(quote_id: int, author_id: int) -> int:
    """Get the rating of a wrong quote."""
    return (await get_wrong_quote(quote_id, author_id)).rating


def get_random_quote_id() -> int:
    """Get random quote id."""
    return random.randint(  # nosec: B311
        1, MAX_QUOTES_ID.value  # type: ignore[attr-defined]
    )


def get_random_author_id() -> int:
    """Get random author id."""
    return random.randint(  # nosec: B311
        1, MAX_AUTHORS_ID.value  # type: ignore[attr-defined]
    )


def get_random_id() -> tuple[int, int]:
    """Get random wrong quote id."""
    return (
        get_random_quote_id(),
        get_random_author_id(),
    )


async def create_wq_and_vote(
    vote: Literal[-1, 1],
    quote_id: int,
    author_id: int,
    contributed_by: str,
    fast: bool = False,
) -> WrongQuote:
    """
    Vote for the wrong_quote with the API.

    If the wrong_quote doesn't exist yet, create it.
    """
    wrong_quote = WRONG_QUOTES_CACHE.get((quote_id, author_id))
    if wrong_quote and wrong_quote.id != -1:
        return await wrong_quote.vote(vote, fast)
    # we don't know the wrong_quote_id, so we have to create the wrong_quote
    wrong_quote = parse_wrong_quote(
        await make_api_request(
            "wrongquotes",
            method="POST",
            body={
                "quote": str(quote_id),
                "author": str(author_id),
                "contributed_by": contributed_by,
            },
        )
    )
    return await wrong_quote.vote(vote, lazy=True)


class QuoteReadyCheckHandler(HTMLRequestHandler):
    """Class that checks if quotes have been loaded."""

    async def check_ready(self) -> None:
        """Fail if quotes aren't ready yet."""
        if not WRONG_QUOTES_CACHE:
            # should work in a few seconds, the quotes just haven't loaded yet
            self.set_header("Retry-After", "5")
            raise HTTPError(503, reason="Service available in a few seconds")

    async def prepare(self) -> None:  # noqa: D102
        await super().prepare()
        if self.request.method != "OPTIONS":
            await self.check_ready()

        if (  # pylint: disable=too-many-boolean-expressions
            self.settings.get("RATELIMITS")
            and self.request.method not in {"HEAD", "OPTIONS"}
            and not self.is_authorized(Permission.RATELIMITS)
            and not self.crawler
            and (
                self.request.path.endswith(".xlsx")
                or self.content_type == "application/vnd.ms-excel"
            )
        ):
            if self.settings.get("UNDER_ATTACK") or not EVENT_REDIS.is_set():
                raise HTTPError(503)

            ratelimited, headers = await ratelimit(
                self.redis,
                self.redis_prefix,
                str(self.request.remote_ip),
                bucket="quotes:image:xlsx",
                max_burst=4,
                count_per_period=1,
                period=60,
                tokens=1 if self.request.method != "HEAD" else 0,
            )

            for header, value in headers.items():
                self.set_header(header, value)

            if ratelimited:
                if self.now.date() == date(self.now.year, 4, 20):
                    self.set_status(420)
                    self.write_error(420)
                else:
                    self.set_status(429)
                    self.write_error(429)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the TraversableStaticFileHandler."""

from __future__ import annotations

from . import (  # noqa: F401  # pylint: disable=unused-import
    FetchCallable,
    app,
    assert_valid_response,
    fetch,
)

CACHE_CONTROL = f"public, immutable, max-age={86400 * 365 * 10}"


async def test_openmoji(fetch: FetchCallable) -> None:  # noqa: F811
    """Test requesting an OpenMoji svg."""
    response = assert_valid_response(
        await fetch("/static/openmoji/svg/1F973.svg?v=15.0.0"),
        content_type="image/svg+xml",
        codes={200},
    )
    assert response.headers["Cache-Control"] == CACHE_CONTROL
    size = len(response.body)

    part_response = assert_valid_response(
        await fetch(
            "/static/openmoji/svg/1F973.svg?v=15.0.0",
            headers={"Range": "bytes=17-41"},
        ),
        content_type="image/svg+xml",
        codes={206},
        needs_to_end_with_line_feed=False,
    )
    assert part_response.headers["Cache-Control"] == CACHE_CONTROL
    assert part_response.headers["Content-Range"] == f"bytes 17-41/{size}"
    assert int(part_response.headers["Content-Length"]) == 25

    part_response = assert_valid_response(
        await fetch(
            "/static/openmoji/svg/1F973.svg?v=15.0.0",
            headers={"Range": "bytes=0-3"},
        ),
        content_type="image/svg+xml",
        codes={206},
        needs_to_end_with_line_feed=False,
    )
    assert part_response.headers["Cache-Control"] == CACHE_CONTROL
    assert part_response.headers["Content-Range"] == f"bytes 0-3/{size}"
    assert int(part_response.headers["Content-Length"]) == 4
    assert part_response.body == b"<svg"

    part_response = assert_valid_response(
        await fetch(
            "/static/openmoji/svg/1F973.svg?v=15.0.0",
            headers={"Range": f"bytes={size - 7}-{size}"},
        ),
        content_type="image/svg+xml",
        codes={206},
    )
    assert part_response.headers["Cache-Control"] == CACHE_CONTROL
    assert (
        part_response.headers["Content-Range"]
        == f"bytes {size - 7}-{size - 1}/{size}"
    )
    assert int(part_response.headers["Content-Length"]) == 7
    assert part_response.body == b"</svg>\n"

    part_response = assert_valid_response(
        await fetch(
            "/static/openmoji/svg/1F973.svg?v=15.0.0",
            headers={"Range": "bytes=-7"},
        ),
        content_type="image/svg+xml",
        codes={206},
    )
    assert part_response.headers["Cache-Control"] == CACHE_CONTROL
    assert (
        part_response.headers["Content-Range"]
        == f"bytes {size - 7}-{size - 1}/{size}"
    )
    assert int(part_response.headers["Content-Length"]) == 7
    assert part_response.body == b"</svg>\n"



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the backdoor."""

from __future__ import annotations

from types import EllipsisType
from typing import Any, Literal, TypedDict, cast

import dill  # type: ignore[import-untyped]  # nosec: B403

from . import (  # noqa: F401  # pylint: disable=unused-import
    FetchCallable,
    app,
    assert_valid_response,
    fetch,
)

Result = tuple[str, Any]  # tuple[str, bytes] before unpickling


class Response(TypedDict):  # noqa: D101
    # pylint: disable=missing-class-docstring
    success: bool | EllipsisType
    output: None | str
    result: None | Result | SystemExit


ErrorTuple = tuple[int, str]


async def request_and_parse(
    # pylint: disable=redefined-outer-name, too-many-arguments
    fetch: FetchCallable,  # noqa: F811
    /,
    code: str,
    *,
    session: None | str = None,
    auth_key: None | str = None,
    mode: Literal["exec", "eval"] = "eval",
    features: None | str = None,
) -> Response | ErrorTuple | str | None:
    """Make a request to the backdoor and parse the response."""
    auth_key = "123qweQWE!@#000000000" if auth_key is None else auth_key
    headers = {"Authorization": auth_key}

    if session:
        headers["X-Backdoor-Session"] = session

    if features:
        headers["X-Future-Feature"] = features

    http_response = await fetch(
        f"/api/backdoor/{mode}",
        method="POST",
        headers=headers,
        body=code,
    )

    assert_valid_response(http_response, "application/vnd.uqfoundation.dill")

    response: Response | ErrorTuple | str | None = dill.loads(  # nosec: B301
        http_response.body
    )

    assert isinstance(response, dict | tuple | str | None)

    if not isinstance(response, dict):
        if isinstance(response, tuple):
            assert len(response) == 2
            assert isinstance(response[0], int)
            assert isinstance(response[1], str)
        return response

    assert len(response) == 3
    assert isinstance(response["success"], bool | EllipsisType)
    assert isinstance(response["output"], None | str)
    assert isinstance(response["result"], None | tuple | SystemExit)
    if isinstance(response["result"], tuple):
        assert len(response["result"]) == 2
        assert isinstance(response["result"][0], str)
        assert isinstance(response["result"][1], bytes | None)
        if isinstance(response["result"][1], bytes):
            result = list(response["result"])
            result[1] = dill.loads(result[1])  # nosec: B301
            response["result"] = cast(Result, tuple(result))

    return response


async def test_backdoor(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the backdoor."""
    # pylint: disable=too-many-statements
    response = await request_and_parse(fetch, "0==0")
    assert isinstance(response, dict)
    assert response["success"]
    assert not response["output"]
    assert response["result"] == ("True", True)

    response = await request_and_parse(
        fetch, "(0<>0)", features="barry_as_FLUFL"
    )
    assert isinstance(response, dict)
    assert response["success"]
    assert not response["output"]
    assert response["result"] == ("False", False)

    response = await request_and_parse(fetch, "print('spam');", mode="eval")
    assert isinstance(response, dict)
    assert not response["success"]
    assert not response["output"]
    assert isinstance(response["result"], tuple)
    assert isinstance(response["result"][1], SyntaxError)

    response = await request_and_parse(fetch, "print('spam');", mode="exec")
    assert isinstance(response, dict)
    assert response["success"]
    assert response["output"] == "spam\n"
    assert not response["result"]

    response = await request_and_parse(
        fetch, "(spam := 'tasty')", session="bacon"
    )
    assert isinstance(response, dict)
    assert response["success"]
    assert not response["output"]
    assert response["result"] == ("'tasty'", "tasty")

    response = await request_and_parse(fetch, "spam", session="bacon")
    assert isinstance(response, dict)
    assert response["success"]
    assert not response["output"]
    assert response["result"] == ("'tasty'", "tasty")

    response = await request_and_parse(fetch, "_", session="bacon")
    assert isinstance(response, dict)
    assert response["success"]
    assert not response["output"]
    assert response["result"] == ("'tasty'", "tasty")

    response = await request_and_parse(fetch, "print")
    assert isinstance(response, dict)
    assert response["success"]
    assert not response["output"]
    assert isinstance(response["result"], tuple)
    assert response["result"][1] is print

    response = await request_and_parse(fetch, "help")
    assert isinstance(response, dict)
    assert response["success"]
    assert not response["output"]
    assert isinstance(response["result"], tuple)
    assert type(response["result"][1]) is type(help)

    response = await request_and_parse(
        fetch, "(await self.load_session())['__name__']"
    )
    assert isinstance(response, dict)
    assert response["success"]
    assert not response["output"]
    assert response["result"] == ("'this'", "this")

    response = await request_and_parse(fetch, "_")
    assert isinstance(response, dict)
    assert not response["success"]
    assert not response["output"]
    assert isinstance(response["result"], tuple)
    assert response["result"][0].startswith("Traceback (most recent call last)")
    assert isinstance(response["result"][1], NameError)

    response = await request_and_parse(
        fetch, "raise SystemExit('spam', 'eggs')", mode="exec"
    )
    assert isinstance(response, dict)
    assert response["success"] is ...
    assert not response["output"]
    assert isinstance(response["result"], SystemExit)
    assert response["result"].args == ("spam", "eggs")

    # # create something that cannot be pickled
    # response = await request_and_parse(
    #     fetch,
    #     "def toast():\n"
    #     "   class Toast: pass\n"
    #     "   return Toast\n"
    #     "LocalToast = toast()\n"
    #     "t = LocalToast()",
    #     mode="exec",
    #     session="sausage",
    # )
    # assert isinstance(response, dict)
    # assert response["success"]
    # assert not response["output"]
    # assert response["result"] is None

    # response = await request_and_parse(
    #     fetch, "raise SystemExit(t)", mode="exec", session="sausage"
    # )
    # assert isinstance(response, dict)
    # assert response["success"] is ...
    # assert not response["output"]
    # assert isinstance(response["result"], SystemExit)
    # assert (
    #     response["result"]
    #     .args[0]
    #     .startswith("<this.toast.<locals>.Toast object at ")
    # )
    # assert response["result"].args[0].endswith(">")

    # response = await request_and_parse(fetch, "t", session="sausage")
    # assert isinstance(response, dict)
    # assert response["success"]
    # assert not response["output"]
    # assert response["result"][0].startswith(
    #     "<this.toast.<locals>.Toast object at "
    # )
    # assert response["result"][0].endswith(">")
    # assert response["result"][1] is None

    response = await request_and_parse(fetch, "0 / 0", mode="exec")
    assert isinstance(response, dict)
    assert not response["success"]
    assert not response["output"]
    assert isinstance(response["result"], tuple)
    assert response["result"][0].startswith("Traceback (most recent call last)")
    assert response["result"][1].args == ("division by zero",)
    assert isinstance(response["result"][1], ZeroDivisionError)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A page that swaps words."""

from __future__ import annotations

import os
from asyncio import Future
from base64 import b64decode, b64encode
from dataclasses import dataclass
from typing import ClassVar, Final

from tornado.web import HTTPError, MissingArgumentError

from ..utils.data_parsing import parse_args
from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from .config_file import InvalidConfigError, SwappedWordsConfig

DIR: Final = os.path.dirname(__file__)

# the max char count of the text to process
MAX_CHAR_COUNT: Final[int] = 32769 - 1

with open(os.path.join(DIR, "config.sw"), encoding="UTF-8") as file:
    DEFAULT_CONFIG: Final[SwappedWordsConfig] = SwappedWordsConfig(file.read())


def check_text_too_long(text: str) -> None:
    """Raise an HTTP error if the text is too long."""
    len_text = len(text)

    if len_text > MAX_CHAR_COUNT:
        raise HTTPError(
            413,
            reason=(
                f"The text has {len_text} characters, but it is only allowed "
                f"to have {MAX_CHAR_COUNT} characters. That's "
                f"{len_text - MAX_CHAR_COUNT} characters too much."
            ),
        )


@dataclass(slots=True)
class SwArgs:
    """Arguments used for swapped words."""

    text: str = ""
    config: None | str = None
    reset: bool = False
    return_config: bool = False
    minify_config: bool = False

    def validate(self) -> None:
        """Validate the given data."""
        self.text = self.text.strip()
        check_text_too_long(self.text)
        if self.config:
            self.config = self.config.strip()

    def validate_require_text(self) -> None:
        """Validate the given data and require the text argument."""
        self.validate()
        if not self.text:
            raise MissingArgumentError("text")


class SwappedWords(HTMLRequestHandler):
    """The request handler for the swapped words page."""

    @parse_args(type_=SwArgs, validation_method="validate")
    async def get(self, *, head: bool = False, args: SwArgs) -> None:
        """Handle GET requests to the swapped words page."""
        # pylint: disable=unused-argument
        await self.handle_text(args)

    def handle_text(self, args: SwArgs) -> Future[None]:
        """Use the text to display the HTML page."""
        if args.config is None:
            cookie = self.get_cookie(
                "swapped-words-config",
                None,
            )
            if cookie is not None:
                # decode the base64 text
                args.config = b64decode(cookie).decode("UTF-8")
        else:
            # save the config in a cookie
            self.set_cookie(
                name="swapped-words-config",
                # encode the config as base64
                value=b64encode(args.config.encode("UTF-8")).decode("UTF-8"),
                expires_days=1000,
                path=self.request.path,
                samesite="Strict",
            )

        try:
            sw_config = (
                DEFAULT_CONFIG
                if args.config is None or args.reset
                else SwappedWordsConfig(args.config)
            )
        except InvalidConfigError as exc:
            self.set_status(400)
            return self.render(
                "pages/swapped_words.html",
                text=args.text,
                output="",
                config=args.config,
                MAX_CHAR_COUNT=MAX_CHAR_COUNT,
                error_msg=str(exc),
            )
        # everything went well
        return self.render(
            "pages/swapped_words.html",
            text=args.text,
            output=sw_config.swap_words(args.text),
            config=sw_config.to_config_str(),
            MAX_CHAR_COUNT=MAX_CHAR_COUNT,
            error_msg=None,
        )

    @parse_args(type_=SwArgs, validation_method="validate_require_text")
    async def post(self, *, args: SwArgs) -> None:
        """Handle POST requests to the swapped words page."""
        await self.handle_text(args)


class SwappedWordsAPI(APIRequestHandler):
    """The request handler for the swapped words API."""

    ALLOWED_METHODS: ClassVar[tuple[str, ...]] = ("GET", "POST")

    @parse_args(type_=SwArgs, validation_method="validate")
    async def get(self, *, head: bool = False, args: SwArgs) -> None:
        """Handle GET requests to the swapped words API."""
        # pylint: disable=unused-argument
        try:
            sw_config = (
                DEFAULT_CONFIG
                if args.config is None
                else SwappedWordsConfig(args.config)
            )

            if args.return_config:
                return await self.finish_dict(
                    text=args.text,
                    return_config=True,
                    minify_config=args.minify_config,
                    config=sw_config.to_config_str(args.minify_config),
                    replaced_text=sw_config.swap_words(args.text),
                )
            return await self.finish_dict(
                text=args.text,
                return_config=False,
                replaced_text=sw_config.swap_words(args.text),
            )
        except InvalidConfigError as exc:
            self.set_status(400)
            return await self.finish_dict(
                error=exc.reason,
                line=exc.line,
                line_num=exc.line_num,
            )

    async def post(self) -> None:
        """Handle POST requests to the swapped words API."""
        return await self.get()  # pylint: disable=missing-kwoa



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Useful stuff for handling static files."""

from __future__ import annotations

import logging
import os
import sys
from collections.abc import Awaitable, Mapping
from functools import cache
from importlib.abc import Traversable
from pathlib import Path
from typing import Any, Final, cast

import defity
import orjson as json
import tornado.web
from blake3 import blake3
from openmoji_dist import VERSION as OPENMOJI_VERSION
from openmoji_dist import get_openmoji_data

from .. import DIR as ROOT_DIR
from .. import STATIC_DIR
from .utils import Handler

LOGGER: Final = logging.getLogger(__name__)


def hash_file(path: str | Path) -> str:
    """Hash a file with BLAKE3."""
    with open(path, "rb") as file:
        return blake3(file.read()).hexdigest(8)


def create_file_hashes_dict() -> dict[str, str]:
    """Create a dict of file hashes."""
    file_hashes_dict = {
        str(path)[len(ROOT_DIR) :].replace(os.path.sep, "/"): hash_file(path)
        for path in Path(STATIC_DIR).rglob("*")
        if path.is_file()
    }
    file_hashes_dict["/favicon.png"] = file_hashes_dict["/static/favicon.png"]
    file_hashes_dict["/favicon.jxl"] = file_hashes_dict["/static/favicon.jxl"]
    file_hashes_dict["/humans.txt"] = file_hashes_dict["/static/humans.txt"]
    return file_hashes_dict


FILE_HASHES_DICT: Final[Mapping[str, str]] = create_file_hashes_dict()

CONTENT_TYPES: Final[Mapping[str, str]] = json.loads(
    Path(ROOT_DIR, "vendored", "media-types.json").read_bytes()
)


def get_handlers() -> list[Handler]:
    """Return a list of handlers for static files."""
    # pylint: disable=import-outside-toplevel
    from .static_file_from_traversable import TraversableStaticFileHandler

    handlers: list[Handler] = [
        (
            "(?i)/static/openmoji/(.*)",
            TraversableStaticFileHandler,
            {"root": get_openmoji_data()},
        ),
        (
            r"(?i)(?:/static)?/(\.env|favicon\.(?:png|jxl)|humans\.txt|robots\.txt)",
            StaticFileHandler,
            {"path": STATIC_DIR},
        ),
    ]
    debug_style_dir = Path(ROOT_DIR).absolute().parent / "style"
    if sys.flags.dev_mode and debug_style_dir.exists():
        # add handlers for the unminified CSS files
        handlers.append(
            (
                r"(?i)/static/css/(.+\.css)",
                StaticFileHandler,
                {"path": str(debug_style_dir)},
            )
        )
    handlers.append(
        (r"(?i)/static/(.*)", CachedStaticFileHandler, {"path": STATIC_DIR})
    )
    return handlers


@cache
def fix_static_path(path: str) -> str:
    """Fix the path for static files."""
    if not path.startswith("/"):
        path = f"/static/{path}"
    if "?" in path:
        path = path.split("?")[0]
    if path.startswith("/static/openmoji/"):
        return f"{path}?v={OPENMOJI_VERSION}"
    path = path.lower()
    if path.startswith("/static/js/utils/"):  # TODO: improve this
        return path
    if path in FILE_HASHES_DICT:
        hash_ = FILE_HASHES_DICT[path]
        return f"{path}?v={hash_}"
    LOGGER.warning("%s not in FILE_HASHES_DICT", path)
    return path


def content_type_from_path(url_path: str, file: Traversable) -> str | None:
    """Extract the Content-Type from a path."""
    content_type: str | None = CONTENT_TYPES.get(Path(url_path).suffix[1:])
    if not content_type:
        with file.open("rb") as io:
            content_type = defity.from_file(io)
    if content_type and content_type.startswith("text/"):
        content_type += "; charset=UTF-8"
    return content_type


class StaticFileHandler(tornado.web.StaticFileHandler):
    """A StaticFileHandler with smart Content-Type."""

    content_type: None | str
    keep_case: bool

    def data_received(self, chunk: bytes) -> None | Awaitable[None]:
        """Do nothing."""

    async def get(self, path: str, include_body: bool = True) -> None:
        """Handle GET requests."""
        if self.keep_case:
            return await super().get(path, include_body=include_body)
        return await super().get(
            path.lower(),
            include_body=include_body,
        )

    async def head(self, path: str) -> None:
        # pylint: disable=invalid-overridden-method
        """Handle HEAD requests."""
        return await self.get(path)

    def initialize(
        self,
        path: str,
        default_filename: None | str = None,
        content_type: None | str = None,
        keep_case: bool = False,
    ) -> None:
        """Initialize the handler."""
        super().initialize(path=path, default_filename=default_filename)
        self.content_type = content_type
        self.keep_case = keep_case

    def set_extra_headers(self, _: str) -> None:
        """Reset the Content-Type header if we know it better."""
        if not self.content_type:
            self.content_type = content_type_from_path(
                self.path, Path(cast(str, self.absolute_path))
            )
        if self.content_type:
            self.set_header("Content-Type", self.content_type)


class CachedStaticFileHandler(StaticFileHandler):
    """A static file handler that sets a smarter Cache-Control header."""

    def compute_etag(self) -> None | str:
        """Don't compute ETag, because it isn't necessary."""
        return None

    @classmethod
    def make_static_path(
        cls, settings: dict[str, Any], path: str, include_version: bool = True
    ) -> str:
        """Make a static path for the given path."""
        # pylint: disable=unused-argument
        return fix_static_path(path)

    def set_headers(self) -> None:
        """Set the default headers for this handler."""
        super().set_headers()
        if not sys.flags.dev_mode and "v" in self.request.arguments:
            self.set_header(  # never changes
                "Cache-Control",
                f"public, immutable, max-age={self.CACHE_MAX_AGE}",
            )



#!/usr/bin/env python3

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# pylint: disable=import-outside-toplevel

"""Nobody inspects the spammish repetition."""

from __future__ import annotations

from importlib.metadata import Distribution
from os import PathLike
from pathlib import Path
from warnings import filterwarnings

from setuptools import setup
from setuptools.build_meta import SetupRequirementsError

BACKEND_REQUIRES = set()
DULWICH = "dulwich==0.21.7"
GET_VERSION = "get_version==3.5.5"
TROVE_CLASSIFIERS = "trove-classifiers==2023.11.29"

filterwarnings("ignore", "", UserWarning, "setuptools.dist")

classifiers = [
    "Development Status :: 5 - Production/Stable",
    (
        "License :: OSI Approved :: "
        "GNU Affero General Public License v3 or later (AGPLv3+)"
    ),
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: Implementation :: CPython",
    "Typing :: Typed",
]


def get_version() -> str:
    """Get the version."""
    if path(".git").exists():
        try:
            # pylint: disable=redefined-outer-name
            from get_version import get_version

            return get_version(__file__, vcs="git")
        except ModuleNotFoundError:
            BACKEND_REQUIRES.add(GET_VERSION)
    return Distribution.at(path(".")).version


def path(path: str | PathLike[str]) -> Path:
    """Return the absolute path to the file."""
    # pylint: disable=redefined-outer-name
    return Path(__file__).resolve().parent / path


if path(".git").exists():
    try:
        from dulwich.repo import Repo
    except ModuleNotFoundError:
        BACKEND_REQUIRES.add(DULWICH)
    else:
        path("REVISION.txt").write_bytes(Repo(path(".")).head())  # type: ignore[arg-type]

    try:
        import trove_classifiers as trove
    except ModuleNotFoundError:
        BACKEND_REQUIRES.add(TROVE_CLASSIFIERS)
    else:
        assert all(_ in trove.classifiers for _ in classifiers)
        assert classifiers == sorted(classifiers)

setup(
    name="an-website",
    license="AGPL-3.0-or-later",
    platforms="OS Independent",
    author="Das Asoziale Netzwerk",
    author_email="contact@asozial.org",
    description="#1 Website in the Worlds",
    long_description_content_type="text/markdown",
    long_description=path("README.md").read_text("UTF-8"),
    version=get_version(),
    url="https://github.com/asozialesnetzwerk/an-website",
    classifiers=classifiers,
    packages=["an_website"],
    python_requires=">=3.10",
    install_requires=path("pip-requirements.txt")
    .read_text("UTF-8")
    .split("\n"),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        "console_scripts": (
            "an-website = an_website.__main__:main",
            "an-backdoor-client = an_website.backdoor.client:main",
        )
    },
)

if BACKEND_REQUIRES:
    raise SetupRequirementsError(BACKEND_REQUIRES)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The utilities package with many helpful things used by other modules."""

from __future__ import annotations

import sys
from os.path import abspath, dirname
from typing import Final

from .request_handler import (
    ElasticRUM,
    ErrorPage,
    NotFoundHandler,
    ZeroDivision,
)
from .utils import ModuleInfo

DIR: Final = abspath(dirname(__file__))


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        name="Utilities",
        description="Nützliche Werkzeuge für alle möglichen Sachen.",
        handlers=(
            (
                r"/error",
                ZeroDivision if sys.flags.dev_mode else NotFoundHandler,
                {},
            ),
            (r"/([1-5][0-9]{2}).html?", ErrorPage, {}),
            (
                r"/@elastic/apm-rum@(5\.12\.0)/dist/bundles"
                r"/elastic-apm-rum.umd(\.min|).js(\.map|)",
                ElasticRUM,
            ),
        ),
        hidden=True,
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for commitment."""

from __future__ import annotations

from datetime import datetime
from os.path import join

from an_website.commitment.commitment import get_commit_data

from . import (  # noqa: F401  # pylint: disable=unused-import
    DIR,
    FetchCallable,
    app,
    assert_valid_json_response,
    assert_valid_response,
    assert_valid_yaml_response,
    fetch,
)


async def test_parsing() -> None:
    """Test parsing the commitment data."""
    file = join(DIR, "commitment.txt")
    for uri in (file, f"file://{file}"):
        data = await get_commit_data(uri)

        assert len(data) == 2

        assert data["50821273052022fbc283e310e09168dc65fb3cce"] == (
            datetime(2022, 8, 29, 19, 56, 6),
            "💬 fix kangaroo comic of today",
        )

        assert data["7335914237808031fa15f32a854ba1e6b1544420"] == (
            datetime(2021, 7, 21, 22, 29, 26),
            "no_js → no_3rd_party",
        )


async def test_text_api(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the commitment API."""
    response = assert_valid_response(
        await fetch("/api/commitment", headers={"Accept": "*/*"}),
        "text/plain;charset=utf-8",
    )

    assert response.body.decode("UTF-8") in {
        "💬 fix kangaroo comic of today\n",
        "no_js → no_3rd_party\n",
    }

    response = assert_valid_response(
        await fetch(
            "/api/commitment?require_emoji=sure", headers={"Accept": "*/*"}
        ),
        "text/plain;charset=utf-8",
    )

    assert response.body.decode("UTF-8") == "💬 fix kangaroo comic of today\n"


async def test_json_api(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the JSON API."""
    for query in (
        "require_emoji=sure",
        "hash=5",
        "hash=50821273052022",
        "hash=50821273052022fbc283e310e09168dc65fb3cce",
    ):
        response = assert_valid_json_response(
            await fetch(
                f"/api/commitment?{query}",
                headers={"Accept": "application/json"},
            ),
        )
        assert response["permalink"].endswith(
            "/api/commitment?hash=50821273052022fbc283e310e09168dc65fb3cce"
        )
        del response["permalink"]
        assert response == {
            "commit_message": "💬 fix kangaroo comic of today",
            "hash": "50821273052022fbc283e310e09168dc65fb3cce",
            "date": "2022-08-29T19:56:06Z",
        }


async def test_yaml_api(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the YAML API."""
    for query in (
        "hash=7",
        "hash=7335914237808031",
        "hash=7335914237808031fa15f32a854ba1e6b1544420",
    ):
        response = assert_valid_yaml_response(
            await fetch(
                f"/api/commitment?{query}",
                headers={"Accept": "application/yaml"},
            ),
        )
        assert response["permalink"].endswith(
            "/api/commitment?hash=7335914237808031fa15f32a854ba1e6b1544420"
        )
        del response["permalink"]
        assert len(response) == 3
        assert response["commit_message"] == "no_js → no_3rd_party"
        assert response["hash"] == "7335914237808031fa15f32a854ba1e6b1544420"
        assert response["date"] in {
            "2021-07-21T22:29:26Z",
            datetime(2021, 7, 21, 22, 29, 26),
        }


async def test_api_404(fetch: FetchCallable) -> None:  # noqa: F811
    """Test getting non-existant hashes."""
    for query in (
        "hash=7335914237808031fa15f32a854ba1e6b154442&require_emoji=sure",
        "hash=9",
        "hash=9168045768942198",
        "hash=9168045768942198e1fbe5e3f00e133b5377a5fa",
        "hash=9168045768942198e1fbe5e3f00e133b5377a5fa6",  # too long
        "hash=x",
        "hash=xxxxxxxxxxxxxxxx",
        "hash=xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "hash=xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",  # too long
    ):
        response = assert_valid_response(
            await fetch(f"/api/commitment?{query}", headers={"Accept": "*/*"}),
            "text/plain;charset=utf-8",
            codes={404},
        )
        assert response.body == b"404 Not Found\n"



# This file is based on the StaticFileHandler from Tornado
# Source: https://github.com/tornadoweb/tornado/blob/b3f2a4bb6fb55f6b1b1e890cdd6332665cfe4a75/tornado/web.py  # noqa: B950  # pylint: disable=line-too-long
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A static file handler for the Traversable abc."""
from __future__ import annotations

import sys
from collections.abc import Awaitable, Iterable
from importlib.abc import Traversable
from pathlib import Path

from tornado import httputil, iostream
from tornado.web import HTTPError, RequestHandler

from .static_file_handling import content_type_from_path


class TraversableStaticFileHandler(RequestHandler):
    """A static file handler for the Traversable abc."""

    root: Traversable

    def data_received(  # noqa: D102
        self, chunk: bytes
    ) -> None | Awaitable[None]:
        pass

    async def get(self, path: str, *, head: bool = False) -> None:  # noqa: C901
        # pylint: disable=too-complex, too-many-branches
        """Handle GET requests for files in the static file directory."""
        if path.startswith("/") or ".." in path.split("/") or "//" in path:
            raise HTTPError(404)

        absolute_path = self.get_absolute_path(path)

        if not absolute_path.is_file():
            raise HTTPError(404)

        self.set_header("Accept-Ranges", "bytes")

        if content_type := self.get_content_type(path, absolute_path):
            self.set_header("Content-Type", content_type)
        del path

        request_range = None
        range_header = self.request.headers.get("Range")
        if range_header:
            # As per RFC 2616 14.16, if an invalid Range header is specified,
            # the request will be treated as if the header didn't exist.
            # pylint: disable-next=protected-access
            request_range = httputil._parse_request_range(range_header)

        size: int = (
            absolute_path.stat().st_size
            if isinstance(absolute_path, Path)
            # pylint: disable-next=bad-builtin
            else sum(map(len, self.get_content(absolute_path)))
        )

        if request_range:
            start, end = request_range
            if start is not None and start < 0:
                start += size
                start = max(start, 0)
            if (
                start is not None
                and (start >= size or (end is not None and start >= end))
            ) or end == 0:  # pylint: disable=use-implicit-booleaness-not-comparison-to-zero # noqa: B950
                # As per RFC 2616 14.35.1, a range is not satisfiable only: if
                # the first requested byte is equal to or greater than the
                # content, or when a suffix with length 0 is specified.
                # https://tools.ietf.org/html/rfc7233#section-2.1
                # A byte-range-spec is invalid if the last-byte-pos value is present
                # and less than the first-byte-pos.
                self.set_status(416)  # Range Not Satisfiable
                self.set_header("Content-Type", "text/plain")
                self.set_header("Content-Range", f"bytes */{size}")
                return
            if end is not None and end > size:
                # Clients sometimes blindly use a large range to limit their
                # download size; cap the endpoint at the actual file size.
                end = size
            # Note: only return HTTP 206 if less than the entire range has been
            # requested. Not only is this semantically correct, but Chrome
            # refuses to play audio if it gets an HTTP 206 in response to
            # ``Range: bytes=0-``.
            if size != (end or size) - (start or 0):
                self.set_status(206)  # Partial Content
                self.set_header(
                    "Content-Range",
                    # pylint: disable-next=protected-access
                    httputil._get_content_range(start, end, size),
                )
        else:
            start = end = None

        content_length = len(range(size)[start:end])
        self.set_header("Content-Length", content_length)

        if head:
            assert self.request.method == "HEAD"
            return
        content = self.get_content(absolute_path, start, end)
        for chunk in content:
            try:
                self.write(chunk)
                await self.flush()
            except iostream.StreamClosedError:
                return

    def get_absolute_path(self, path: str) -> Traversable:
        """Get the absolute path of a file."""
        return self.root / path

    @classmethod
    def get_content(
        cls,
        abspath: Traversable,
        start: int | None = None,
        end: int | None = None,
    ) -> Iterable[bytes]:
        """Read the content of a file in chunks."""
        with abspath.open("rb") as file:
            if start is not None:
                file.seek(start)
            remaining: int | None = (
                (end - (start or 0)) if end is not None else None
            )

            while True:  # pylint: disable=while-used
                chunk_size = 64 * 1024
                if remaining is not None and remaining < chunk_size:
                    chunk_size = remaining
                chunk = file.read(chunk_size)
                if chunk:
                    if remaining is not None:
                        remaining -= len(chunk)
                    yield chunk
                else:
                    assert not remaining
                    return

    @classmethod
    def get_content_type(
        cls, path: str, absolute_path: Traversable
    ) -> str | None:
        """Get the content-type of a file."""
        return content_type_from_path(path, absolute_path)

    def head(self, path: str) -> Awaitable[None]:
        """Handle HEAD requests for files in the static file directory."""
        return self.get(path, head=True)

    def initialize(self, root: Traversable) -> None:
        """Initialize this handler with a root directory."""
        self.root = root

    def set_default_headers(self) -> None:
        """Set the default headers for this handler."""
        super().set_default_headers()
        if not sys.flags.dev_mode and "v" in self.request.arguments:
            self.set_header(  # never changes
                "Cache-Control",
                f"public, immutable, max-age={86400 * 365 * 10}",
            )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The page that generates art."""

from __future__ import annotations

from typing import ClassVar

from redis.asyncio import Redis
from tornado.web import HTTPError

from .. import EVENT_REDIS
from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import ModuleInfo


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (r"/LOLWUT", LOLWUT),
            (r"/LOLWUT/((?:\d+/)*\d+)", LOLWUT),
            (r"/API/LOLWUT", LOLWUTAPI),
            (r"/API/LOLWUT/((?:\d+/)*\d+)", LOLWUTAPI),
        ),
        name="LOLWUT",
        description="LOLWUT; präsentiert von Redis",
        path="/LOLWUT",
        keywords=(
            "LOLWUT",
            "Redis",
        ),
    )


async def generate_art(
    redis: Redis[str],
    args: None | str = None,
    head: bool = False,
) -> None | str:
    """Generate art."""
    if args:
        arguments = args.split("/")
        for argument in arguments:
            if not argument:
                raise HTTPError(404)
        command = "LOLWUT VERSION " + " ".join(arguments)
    else:
        command = "LOLWUT"
    if head:
        return None
    return await redis.execute_command(command)  # type: ignore[no-any-return, no-untyped-call]  # noqa: B950


class LOLWUT(HTMLRequestHandler):
    """The request handler for the LOLWUT page."""

    async def get(self, args: None | str = None, *, head: bool = False) -> None:
        """Handle GET requests to the LOLWUT page."""
        if not EVENT_REDIS.is_set():
            raise HTTPError(503)

        art = await generate_art(self.redis, args, head)

        if head:
            return

        await self.render(
            "ansi2html.html",
            ansi=art,
            powered_by="https://redis.io",
            powered_by_name="Redis",
        )


class LOLWUTAPI(APIRequestHandler):
    """The request handler for the LOLWUT API."""

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        "text/plain",
        *APIRequestHandler.POSSIBLE_CONTENT_TYPES,
    )

    async def get(self, args: None | str = None, *, head: bool = False) -> None:
        """Handle GET requests to the LOLWUT API."""
        if not EVENT_REDIS.is_set():
            raise HTTPError(503)

        art = await generate_art(self.redis, args, head)

        if self.content_type == "text/plain":
            return await self.finish(art)

        await self.finish({"LOLWUT": art})
